import typing

from . import (
    anchor_decimal,
    asset,
    cross_margin_account_info,
    expire_series_override_args,
    expiry_series,
    expiry_series_status,
    halt_args,
    halt_state,
    halt_state_args,
    halt_state_v2,
    initialize_market_args,
    initialize_market_node_args,
    initialize_state_args,
    initialize_zeta_group_args,
    initialize_zeta_pricing_args,
    kind,
    margin_account_type,
    margin_parameters,
    margin_requirement,
    movement_type,
    order_args,
    order_complete_type,
    order_state,
    order_type,
    override_expiry_args,
    perp_parameters,
    place_order_type,
    position,
    position_movement_arg,
    pricing_parameters,
    product,
    product_greeks,
    product_ledger,
    side,
    strike,
    trait_type,
    treasury_movement_type,
    trigger_direction,
    update_greeks_args,
    update_interest_rate_args,
    update_margin_parameters_args,
    update_perp_parameters_args,
    update_pricing_parameters_args,
    update_state_args,
    update_volatility_args,
    update_zeta_group_expiry_args,
    update_zeta_pricing_pubkeys_args,
    validation_type,
)
from .anchor_decimal import AnchorDecimal, AnchorDecimalJSON
from .asset import AssetJSON, AssetKind
from .cross_margin_account_info import (
    CrossMarginAccountInfo,
    CrossMarginAccountInfoJSON,
)
from .expire_series_override_args import (
    ExpireSeriesOverrideArgs,
    ExpireSeriesOverrideArgsJSON,
)
from .expiry_series import ExpirySeries, ExpirySeriesJSON
from .expiry_series_status import ExpirySeriesStatusJSON, ExpirySeriesStatusKind
from .halt_args import HaltArgs, HaltArgsJSON
from .halt_state import HaltState, HaltStateJSON
from .halt_state_args import HaltStateArgs, HaltStateArgsJSON
from .halt_state_v2 import HaltStateV2, HaltStateV2JSON
from .initialize_market_args import InitializeMarketArgs, InitializeMarketArgsJSON
from .initialize_market_node_args import (
    InitializeMarketNodeArgs,
    InitializeMarketNodeArgsJSON,
)
from .initialize_state_args import InitializeStateArgs, InitializeStateArgsJSON
from .initialize_zeta_group_args import (
    InitializeZetaGroupArgs,
    InitializeZetaGroupArgsJSON,
)
from .initialize_zeta_pricing_args import (
    InitializeZetaPricingArgs,
    InitializeZetaPricingArgsJSON,
)
from .kind import KindJSON, KindKind
from .margin_account_type import MarginAccountTypeJSON, MarginAccountTypeKind
from .margin_parameters import MarginParameters, MarginParametersJSON
from .margin_requirement import MarginRequirementJSON, MarginRequirementKind
from .movement_type import MovementTypeJSON, MovementTypeKind
from .order_args import OrderArgs, OrderArgsJSON
from .order_complete_type import OrderCompleteTypeJSON, OrderCompleteTypeKind
from .order_state import OrderState, OrderStateJSON
from .order_type import OrderTypeJSON, OrderTypeKind
from .override_expiry_args import OverrideExpiryArgs, OverrideExpiryArgsJSON
from .perp_parameters import PerpParameters, PerpParametersJSON
from .place_order_type import PlaceOrderTypeJSON, PlaceOrderTypeKind
from .position import Position, PositionJSON
from .position_movement_arg import PositionMovementArg, PositionMovementArgJSON
from .pricing_parameters import PricingParameters, PricingParametersJSON
from .product import Product, ProductJSON
from .product_greeks import ProductGreeks, ProductGreeksJSON
from .product_ledger import ProductLedger, ProductLedgerJSON
from .self_trade_behavior_zeta import (
    SelfTradeBehaviorZetaJSON,
    SelfTradeBehaviorZetaKind,
)
from .side import SideJSON, SideKind
from .strike import Strike, StrikeJSON
from .trait_type import TraitTypeJSON, TraitTypeKind
from .treasury_movement_type import TreasuryMovementTypeJSON, TreasuryMovementTypeKind
from .trigger_direction import TriggerDirectionJSON, TriggerDirectionKind
from .update_greeks_args import UpdateGreeksArgs, UpdateGreeksArgsJSON
from .update_interest_rate_args import (
    UpdateInterestRateArgs,
    UpdateInterestRateArgsJSON,
)
from .update_margin_parameters_args import (
    UpdateMarginParametersArgs,
    UpdateMarginParametersArgsJSON,
)
from .update_perp_parameters_args import (
    UpdatePerpParametersArgs,
    UpdatePerpParametersArgsJSON,
)
from .update_pricing_parameters_args import (
    UpdatePricingParametersArgs,
    UpdatePricingParametersArgsJSON,
)
from .update_state_args import UpdateStateArgs, UpdateStateArgsJSON
from .update_volatility_args import UpdateVolatilityArgs, UpdateVolatilityArgsJSON
from .update_zeta_group_expiry_args import (
    UpdateZetaGroupExpiryArgs,
    UpdateZetaGroupExpiryArgsJSON,
)
from .update_zeta_pricing_pubkeys_args import (
    UpdateZetaPricingPubkeysArgs,
    UpdateZetaPricingPubkeysArgsJSON,
)
from .validation_type import ValidationTypeJSON, ValidationTypeKind



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey
from spl.token.constants import TOKEN_PROGRAM_ID

from ..program_id import PROGRAM_ID


class DepositV2Args(typing.TypedDict):
    amount: int


layout = borsh.CStruct("amount" / borsh.U64)


class DepositV2Accounts(typing.TypedDict):
    margin_account: Pubkey
    vault: Pubkey
    user_token_account: Pubkey
    socialized_loss_account: Pubkey
    authority: Pubkey
    state: Pubkey
    pricing: Pubkey


def deposit_v2(
    args: DepositV2Args,
    accounts: DepositV2Accounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["margin_account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["vault"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["user_token_account"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["socialized_loss_account"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(pubkey=accounts["authority"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["pricing"], is_signer=False, is_writable=True),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"mKE\x99\xac\xda\x92\x13"
    encoded_args = layout.build(
        {
            "amount": args["amount"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing

from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from ..program_id import PROGRAM_ID


class AddMarketIndexesAccounts(typing.TypedDict):
    market_indexes: Pubkey
    zeta_group: Pubkey


def add_market_indexes(
    accounts: AddMarketIndexesAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["market_indexes"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["zeta_group"], is_signer=False, is_writable=True),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"^\xf6\x90\xaf\x04\xa4\xe9\xfc"
    encoded_args = b""
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import BorshPubkey
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
from solders.pubkey import Pubkey

from .. import types
from ..program_id import PROGRAM_ID


class CrossMarginAccountManagerJSON(typing.TypedDict):
    nonce: int
    authority: str
    accounts: list[types.cross_margin_account_info.CrossMarginAccountInfoJSON]
    referrer: str
    airdrop_community: int
    padding: list[int]


@dataclass
class CrossMarginAccountManager:
    discriminator: typing.ClassVar = b"\\\xa2\x1aC1V\xfc\x05"
    layout: typing.ClassVar = borsh.CStruct(
        "nonce" / borsh.U8,
        "authority" / BorshPubkey,
        "accounts" / types.cross_margin_account_info.CrossMarginAccountInfo.layout[20],
        "referrer" / BorshPubkey,
        "airdrop_community" / borsh.U8,
        "padding" / borsh.U8[22],
    )
    nonce: int
    authority: Pubkey
    accounts: list[types.cross_margin_account_info.CrossMarginAccountInfo]
    referrer: Pubkey
    airdrop_community: int
    padding: list[int]

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["CrossMarginAccountManager"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["CrossMarginAccountManager"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["CrossMarginAccountManager"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "CrossMarginAccountManager":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator("The discriminator for this account is invalid")
        dec = CrossMarginAccountManager.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            nonce=dec.nonce,
            authority=dec.authority,
            accounts=list(
                map(
                    lambda item: types.cross_margin_account_info.CrossMarginAccountInfo.from_decoded(item),
                    dec.accounts,
                )
            ),
            referrer=dec.referrer,
            airdrop_community=dec.airdrop_community,
            padding=dec.padding,
        )

    def to_json(self) -> CrossMarginAccountManagerJSON:
        return {
            "nonce": self.nonce,
            "authority": str(self.authority),
            "accounts": list(map(lambda item: item.to_json(), self.accounts)),
            "referrer": str(self.referrer),
            "airdrop_community": self.airdrop_community,
            "padding": self.padding,
        }

    @classmethod
    def from_json(cls, obj: CrossMarginAccountManagerJSON) -> "CrossMarginAccountManager":
        return cls(
            nonce=obj["nonce"],
            authority=Pubkey.from_string(obj["authority"]),
            accounts=list(
                map(
                    lambda item: types.cross_margin_account_info.CrossMarginAccountInfo.from_json(item),
                    obj["accounts"],
                )
            ),
            referrer=Pubkey.from_string(obj["referrer"]),
            airdrop_community=obj["airdrop_community"],
            padding=obj["padding"],
        )



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from .. import types
from ..program_id import PROGRAM_ID


class UpdatePerpParametersArgs(typing.TypedDict):
    args: types.update_perp_parameters_args.UpdatePerpParametersArgs
    asset: types.asset.AssetKind


layout = borsh.CStruct(
    "args" / types.update_perp_parameters_args.UpdatePerpParametersArgs.layout,
    "asset" / types.asset.layout,
)


class UpdatePerpParametersAccounts(typing.TypedDict):
    state: Pubkey
    pricing: Pubkey
    admin: Pubkey


def update_perp_parameters(
    args: UpdatePerpParametersArgs,
    accounts: UpdatePerpParametersAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["pricing"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"Z\x87\xdb*\xa4\x86a\xae"
    encoded_args = layout.build(
        {
            "args": args["args"].to_encodable(),
            "asset": args["asset"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from .. import types
from ..program_id import PROGRAM_ID


class UpdateZetaGroupMarginParametersArgs(typing.TypedDict):
    args: types.update_margin_parameters_args.UpdateMarginParametersArgs


layout = borsh.CStruct("args" / types.update_margin_parameters_args.UpdateMarginParametersArgs.layout)


class UpdateZetaGroupMarginParametersAccounts(typing.TypedDict):
    state: Pubkey
    zeta_group: Pubkey
    admin: Pubkey


def update_zeta_group_margin_parameters(
    args: UpdateZetaGroupMarginParametersArgs,
    accounts: UpdateZetaGroupMarginParametersAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["zeta_group"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"<\xd0y\x93\xf2j\x0b\xfe"
    encoded_args = layout.build(
        {
            "args": args["args"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import BorshPubkey
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
from solders.pubkey import Pubkey

from .. import types
from ..program_id import PROGRAM_ID


class MarketStateJSON(typing.TypedDict):
    account_flags: types.account_flags.AccountFlagsJSON
    own_address: str
    vault_signer_nonce: int
    base_mint: str
    quote_mint: str
    base_vault: str
    base_deposits_total: int
    base_fees_accrued: int
    quote_vault: str
    quote_deposits_total: int
    quote_fees_accrued: int
    quote_dust_threshold: int
    request_queue: str
    event_queue: str
    bids: str
    asks: str
    base_lot_size: int
    quote_lot_size: int
    fee_rate_bps: int
    referrer_rebate_accrued: int
    open_orders_authority: str
    prune_authority: str
    consume_events_authority: str
    epoch_length: int
    epoch_start_ts: int
    start_epoch_seq_num: int


@dataclass
class MarketState:
    discriminator: typing.ClassVar = b"\x73\x65\x72\x75\x6D"
    layout: typing.ClassVar = borsh.CStruct(
        "account_flags" / types.account_flags.AccountFlags.layout,
        "own_address" / BorshPubkey,
        "vault_signer_nonce" / borsh.U64,
        "base_mint" / BorshPubkey,
        "quote_mint" / BorshPubkey,
        "base_vault" / BorshPubkey,
        "base_deposits_total" / borsh.U64,
        "base_fees_accrued" / borsh.U64,
        "quote_vault" / BorshPubkey,
        "quote_deposits_total" / borsh.U64,
        "quote_fees_accrued" / borsh.U64,
        "quote_dust_threshold" / borsh.U64,
        "request_queue" / BorshPubkey,
        "event_queue" / BorshPubkey,
        "bids" / BorshPubkey,
        "asks" / BorshPubkey,
        "base_lot_size" / borsh.U64,
        "quote_lot_size" / borsh.U64,
        "fee_rate_bps" / borsh.U64,
        "referrer_rebate_accrued" / borsh.U64,
        "open_orders_authority" / BorshPubkey,
        "prune_authority" / BorshPubkey,
        "consume_events_authority" / BorshPubkey,
        "epoch_length" / borsh.U16,
        "epoch_start_ts" / borsh.U64,
        "start_epoch_seq_num" / borsh.U64,
        "padding" / borsh.U8[981],  # padding
    )
    account_flags: types.account_flags.AccountFlags
    own_address: Pubkey
    vault_signer_nonce: int
    base_mint: Pubkey
    quote_mint: Pubkey
    base_vault: Pubkey
    base_deposits_total: int
    base_fees_accrued: int
    quote_vault: Pubkey
    quote_deposits_total: int
    quote_fees_accrued: int
    quote_dust_threshold: int
    request_queue: Pubkey
    event_queue: Pubkey
    bids: Pubkey
    asks: Pubkey
    base_lot_size: int
    quote_lot_size: int
    fee_rate_bps: int
    referrer_rebate_accrued: int
    open_orders_authority: Pubkey
    prune_authority: Pubkey
    consume_events_authority: Pubkey
    epoch_length: int
    epoch_start_ts: int
    start_epoch_seq_num: int

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["MarketState"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["MarketState"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["MarketState"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "MarketState":
        if data[: len(cls.discriminator)] != cls.discriminator:
            raise AccountInvalidDiscriminator("The discriminator for this account is invalid")
        dec = MarketState.layout.parse(data[len(cls.discriminator) :])
        return cls(
            account_flags=types.account_flags.AccountFlags.from_decoded(dec.account_flags),
            own_address=dec.own_address,
            vault_signer_nonce=dec.vault_signer_nonce,
            base_mint=dec.base_mint,
            quote_mint=dec.quote_mint,
            base_vault=dec.base_vault,
            base_deposits_total=dec.base_deposits_total,
            base_fees_accrued=dec.base_fees_accrued,
            quote_vault=dec.quote_vault,
            quote_deposits_total=dec.quote_deposits_total,
            quote_fees_accrued=dec.quote_fees_accrued,
            quote_dust_threshold=dec.quote_dust_threshold,
            request_queue=dec.request_queue,
            event_queue=dec.event_queue,
            bids=dec.bids,
            asks=dec.asks,
            base_lot_size=dec.base_lot_size,
            quote_lot_size=dec.quote_lot_size,
            fee_rate_bps=dec.fee_rate_bps,
            referrer_rebate_accrued=dec.referrer_rebate_accrued,
            open_orders_authority=dec.open_orders_authority,
            prune_authority=dec.prune_authority,
            consume_events_authority=dec.consume_events_authority,
            epoch_length=dec.epoch_length,
            epoch_start_ts=dec.epoch_start_ts,
            start_epoch_seq_num=dec.start_epoch_seq_num,
        )



import typing

from anchorpy.error import ProgramError


class DepositOverflow(ProgramError):
    def __init__(self) -> None:
        super().__init__(6000, "Deposit overflow")

    code = 6000
    name = "DepositOverflow"
    msg = "Deposit overflow"


class Unreachable(ProgramError):
    def __init__(self) -> None:
        super().__init__(6001, "Unreachable")

    code = 6001
    name = "Unreachable"
    msg = "Unreachable"


class FailedInitialMarginRequirement(ProgramError):
    def __init__(self) -> None:
        super().__init__(6002, "Failed initial margin requirement")

    code = 6002
    name = "FailedInitialMarginRequirement"
    msg = "Failed initial margin requirement"


class LiquidatorFailedMarginRequirement(ProgramError):
    def __init__(self) -> None:
        super().__init__(6003, "Liquidator failed margin requirement")

    code = 6003
    name = "LiquidatorFailedMarginRequirement"
    msg = "Liquidator failed margin requirement"


class CannotLiquidateOwnAccount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6004, "Cannot liquidate own account")

    code = 6004
    name = "CannotLiquidateOwnAccount"
    msg = "Cannot liquidate own account"


class CrankInvalidRemainingAccounts(ProgramError):
    def __init__(self) -> None:
        super().__init__(6005, "Invalid cranking remaining accounts")

    code = 6005
    name = "CrankInvalidRemainingAccounts"
    msg = "Invalid cranking remaining accounts"


class IncorrectTickSize(ProgramError):
    def __init__(self) -> None:
        super().__init__(6006, "Incorrect tick size")

    code = 6006
    name = "IncorrectTickSize"
    msg = "Incorrect tick size"


class ZeroPrice(ProgramError):
    def __init__(self) -> None:
        super().__init__(6007, "ZeroPrice")

    code = 6007
    name = "ZeroPrice"
    msg = "ZeroPrice"


class ZeroSize(ProgramError):
    def __init__(self) -> None:
        super().__init__(6008, "ZeroSize")

    code = 6008
    name = "ZeroSize"
    msg = "ZeroSize"


class ZeroWithdrawableBalance(ProgramError):
    def __init__(self) -> None:
        super().__init__(6009, "Zero withdrawable balance")

    code = 6009
    name = "ZeroWithdrawableBalance"
    msg = "Zero withdrawable balance"


class DepositAmountExceeded(ProgramError):
    def __init__(self) -> None:
        super().__init__(6010, "Deposit amount exceeds limit and user is not whitelisted")

    code = 6010
    name = "DepositAmountExceeded"
    msg = "Deposit amount exceeds limit and user is not whitelisted"


class WithdrawalAmountExceedsWithdrawableBalance(ProgramError):
    def __init__(self) -> None:
        super().__init__(6011, "Withdrawal amount exceeds withdrawable balance")

    code = 6011
    name = "WithdrawalAmountExceedsWithdrawableBalance"
    msg = "Withdrawal amount exceeds withdrawable balance"


class AccountHasSufficientMarginPostCancels(ProgramError):
    def __init__(self) -> None:
        super().__init__(6012, "Account has sufficient margin post cancels")

    code = 6012
    name = "AccountHasSufficientMarginPostCancels"
    msg = "Account has sufficient margin post cancels"


class OverBankrupt(ProgramError):
    def __init__(self) -> None:
        super().__init__(6013, "Over bankrupt")

    code = 6013
    name = "OverBankrupt"
    msg = "Over bankrupt"


class AccountHasSufficientMargin(ProgramError):
    def __init__(self) -> None:
        super().__init__(6014, "Account has sufficient margin")

    code = 6014
    name = "AccountHasSufficientMargin"
    msg = "Account has sufficient margin"


class UserHasNoActiveOrders(ProgramError):
    def __init__(self) -> None:
        super().__init__(6015, "User has no active orders")

    code = 6015
    name = "UserHasNoActiveOrders"
    msg = "User has no active orders"


class InvalidExpirationInterval(ProgramError):
    def __init__(self) -> None:
        super().__init__(6016, "Invalid expiration interval")

    code = 6016
    name = "InvalidExpirationInterval"
    msg = "Invalid expiration interval"


class ProductMarketsAlreadyInitialized(ProgramError):
    def __init__(self) -> None:
        super().__init__(6017, "Product markets already initialized")

    code = 6017
    name = "ProductMarketsAlreadyInitialized"
    msg = "Product markets already initialized"


class InvalidProductMarketKey(ProgramError):
    def __init__(self) -> None:
        super().__init__(6018, "Invalid product market key")

    code = 6018
    name = "InvalidProductMarketKey"
    msg = "Invalid product market key"


class MarketNotLive(ProgramError):
    def __init__(self) -> None:
        super().__init__(6019, "Market not live")

    code = 6019
    name = "MarketNotLive"
    msg = "Market not live"


class MarketPricingNotReady(ProgramError):
    def __init__(self) -> None:
        super().__init__(6020, "Market pricing not ready")

    code = 6020
    name = "MarketPricingNotReady"
    msg = "Market pricing not ready"


class UserHasRemainingOrdersOnExpiredMarket(ProgramError):
    def __init__(self) -> None:
        super().__init__(6021, "User has remaining orders on expired market")

    code = 6021
    name = "UserHasRemainingOrdersOnExpiredMarket"
    msg = "User has remaining orders on expired market"


class InvalidSeriesExpiration(ProgramError):
    def __init__(self) -> None:
        super().__init__(6022, "Invalid series expiration")

    code = 6022
    name = "InvalidSeriesExpiration"
    msg = "Invalid series expiration"


class InvalidExpiredOrderCancel(ProgramError):
    def __init__(self) -> None:
        super().__init__(6023, "Invalid expired order cancel")

    code = 6023
    name = "InvalidExpiredOrderCancel"
    msg = "Invalid expired order cancel"


class NoMarketsToAdd(ProgramError):
    def __init__(self) -> None:
        super().__init__(6024, "No markets to add")

    code = 6024
    name = "NoMarketsToAdd"
    msg = "No markets to add"


class UserHasUnsettledPositions(ProgramError):
    def __init__(self) -> None:
        super().__init__(6025, "User has unsettled positions")

    code = 6025
    name = "UserHasUnsettledPositions"
    msg = "User has unsettled positions"


class NoMarginAccountsToSettle(ProgramError):
    def __init__(self) -> None:
        super().__init__(6026, "No margin accounts to settle")

    code = 6026
    name = "NoMarginAccountsToSettle"
    msg = "No margin accounts to settle"


class CannotSettleUserWithActiveOrders(ProgramError):
    def __init__(self) -> None:
        super().__init__(6027, "Cannot settle users with active orders")

    code = 6027
    name = "CannotSettleUserWithActiveOrders"
    msg = "Cannot settle users with active orders"


class OrderbookNotEmpty(ProgramError):
    def __init__(self) -> None:
        super().__init__(6028, "Orderbook not empty")

    code = 6028
    name = "OrderbookNotEmpty"
    msg = "Orderbook not empty"


class InvalidNumberOfAccounts(ProgramError):
    def __init__(self) -> None:
        super().__init__(6029, "Invalid number of accounts")

    code = 6029
    name = "InvalidNumberOfAccounts"
    msg = "Invalid number of accounts"


class InvalidMarketAccounts(ProgramError):
    def __init__(self) -> None:
        super().__init__(6030, "Bids or Asks don't match the Market")

    code = 6030
    name = "InvalidMarketAccounts"
    msg = "Bids or Asks don't match the Market"


class ProductStrikeUninitialized(ProgramError):
    def __init__(self) -> None:
        super().__init__(6031, "Product strike uninitialized")

    code = 6031
    name = "ProductStrikeUninitialized"
    msg = "Product strike uninitialized"


class PricingNotUpToDate(ProgramError):
    def __init__(self) -> None:
        super().__init__(6032, "Pricing not up to date")

    code = 6032
    name = "PricingNotUpToDate"
    msg = "Pricing not up to date"


class RetreatsAreStale(ProgramError):
    def __init__(self) -> None:
        super().__init__(6033, "Retreats are stale")

    code = 6033
    name = "RetreatsAreStale"
    msg = "Retreats are stale"


class ProductDirty(ProgramError):
    def __init__(self) -> None:
        super().__init__(6034, "Product dirty")

    code = 6034
    name = "ProductDirty"
    msg = "Product dirty"


class ProductStrikesInitialized(ProgramError):
    def __init__(self) -> None:
        super().__init__(6035, "Product strikes initialized")

    code = 6035
    name = "ProductStrikesInitialized"
    msg = "Product strikes initialized"


class StrikeInitializationNotReady(ProgramError):
    def __init__(self) -> None:
        super().__init__(6036, "Strike initialization not ready")

    code = 6036
    name = "StrikeInitializationNotReady"
    msg = "Strike initialization not ready"


class UnsupportedKind(ProgramError):
    def __init__(self) -> None:
        super().__init__(6037, "Unsupported kind")

    code = 6037
    name = "UnsupportedKind"
    msg = "Unsupported kind"


class InvalidZetaGroup(ProgramError):
    def __init__(self) -> None:
        super().__init__(6038, "Invalid zeta group")

    code = 6038
    name = "InvalidZetaGroup"
    msg = "Invalid zeta group"


class InvalidMarginAccount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6039, "Invalid margin account")

    code = 6039
    name = "InvalidMarginAccount"
    msg = "Invalid margin account"


class InvalidGreeksAccount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6040, "Invalid greeks account")

    code = 6040
    name = "InvalidGreeksAccount"
    msg = "Invalid greeks account"


class InvalidSettlementAccount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6041, "Invalid settlement account")

    code = 6041
    name = "InvalidSettlementAccount"
    msg = "Invalid settlement account"


class InvalidCancelAuthority(ProgramError):
    def __init__(self) -> None:
        super().__init__(6042, "Invalid cancel authority")

    code = 6042
    name = "InvalidCancelAuthority"
    msg = "Invalid cancel authority"


class CannotUpdatePricingAfterExpiry(ProgramError):
    def __init__(self) -> None:
        super().__init__(6043, "Cannot update pricing after expiry")

    code = 6043
    name = "CannotUpdatePricingAfterExpiry"
    msg = "Cannot update pricing after expiry"


class LoadAccountDiscriminatorAlreadySet(ProgramError):
    def __init__(self) -> None:
        super().__init__(6044, "Account discriminator already set")

    code = 6044
    name = "LoadAccountDiscriminatorAlreadySet"
    msg = "Account discriminator already set"


class AccountAlreadyInitialized(ProgramError):
    def __init__(self) -> None:
        super().__init__(6045, "Account already initialized")

    code = 6045
    name = "AccountAlreadyInitialized"
    msg = "Account already initialized"


class GreeksAccountSeedsMismatch(ProgramError):
    def __init__(self) -> None:
        super().__init__(6046, "Greeks account seeds mismatch")

    code = 6046
    name = "GreeksAccountSeedsMismatch"
    msg = "Greeks account seeds mismatch"


class ZetaGroupAccountSeedsMismatch(ProgramError):
    def __init__(self) -> None:
        super().__init__(6047, "Zeta group account seeds mismatch")

    code = 6047
    name = "ZetaGroupAccountSeedsMismatch"
    msg = "Zeta group account seeds mismatch"


class MarginAccountSeedsMismatch(ProgramError):
    def __init__(self) -> None:
        super().__init__(6048, "Margin account seeds mismatch")

    code = 6048
    name = "MarginAccountSeedsMismatch"
    msg = "Margin account seeds mismatch"


class OpenOrdersAccountSeedsMismatch(ProgramError):
    def __init__(self) -> None:
        super().__init__(6049, "Open orders account seeds mismatch")

    code = 6049
    name = "OpenOrdersAccountSeedsMismatch"
    msg = "Open orders account seeds mismatch"


class MarketNodeAccountSeedsMismatch(ProgramError):
    def __init__(self) -> None:
        super().__init__(6050, "Market node seeds mismatch")

    code = 6050
    name = "MarketNodeAccountSeedsMismatch"
    msg = "Market node seeds mismatch"


class UserTradingFeeWhitelistAccountSeedsMismatch(ProgramError):
    def __init__(self) -> None:
        super().__init__(6051, "User trading fee whitelist account seeds mismatch")

    code = 6051
    name = "UserTradingFeeWhitelistAccountSeedsMismatch"
    msg = "User trading fee whitelist account seeds mismatch"


class UserDepositWhitelistAccountSeedsMismatch(ProgramError):
    def __init__(self) -> None:
        super().__init__(6052, "User deposit whitelist account seeds mismatch")

    code = 6052
    name = "UserDepositWhitelistAccountSeedsMismatch"
    msg = "User deposit whitelist account seeds mismatch"


class MarketIndexesUninitialized(ProgramError):
    def __init__(self) -> None:
        super().__init__(6053, "Market indexes uninitialized")

    code = 6053
    name = "MarketIndexesUninitialized"
    msg = "Market indexes uninitialized"


class MarketIndexesAlreadyInitialized(ProgramError):
    def __init__(self) -> None:
        super().__init__(6054, "Market indexes already initialized")

    code = 6054
    name = "MarketIndexesAlreadyInitialized"
    msg = "Market indexes already initialized"


class CannotGetUnsetStrike(ProgramError):
    def __init__(self) -> None:
        super().__init__(6055, "Cannot get unset strike")

    code = 6055
    name = "CannotGetUnsetStrike"
    msg = "Cannot get unset strike"


class CannotSetInitializedStrike(ProgramError):
    def __init__(self) -> None:
        super().__init__(6056, "Cannot set initialized strike")

    code = 6056
    name = "CannotSetInitializedStrike"
    msg = "Cannot set initialized strike"


class CannotResetUninitializedStrike(ProgramError):
    def __init__(self) -> None:
        super().__init__(6057, "Cannot set initialized strike")

    code = 6057
    name = "CannotResetUninitializedStrike"
    msg = "Cannot set initialized strike"


class CrankMarginAccountNotMutable(ProgramError):
    def __init__(self) -> None:
        super().__init__(6058, "CrankMarginAccountNotMutable")

    code = 6058
    name = "CrankMarginAccountNotMutable"
    msg = "CrankMarginAccountNotMutable"


class InvalidAdminSigner(ProgramError):
    def __init__(self) -> None:
        super().__init__(6059, "InvalidAdminSigner")

    code = 6059
    name = "InvalidAdminSigner"
    msg = "InvalidAdminSigner"


class UserHasActiveOrders(ProgramError):
    def __init__(self) -> None:
        super().__init__(6060, "User still has active orders")

    code = 6060
    name = "UserHasActiveOrders"
    msg = "User still has active orders"


class UserForceCancelInProgress(ProgramError):
    def __init__(self) -> None:
        super().__init__(6061, "User has a force cancel in progress")

    code = 6061
    name = "UserForceCancelInProgress"
    msg = "User has a force cancel in progress"


class FailedPriceBandCheck(ProgramError):
    def __init__(self) -> None:
        super().__init__(6062, "Failed price band check")

    code = 6062
    name = "FailedPriceBandCheck"
    msg = "Failed price band check"


class UnsortedOpenOrdersAccounts(ProgramError):
    def __init__(self) -> None:
        super().__init__(6063, "Unsorted open orders accounts")

    code = 6063
    name = "UnsortedOpenOrdersAccounts"
    msg = "Unsorted open orders accounts"


class AccountNotMutable(ProgramError):
    def __init__(self) -> None:
        super().__init__(6064, "Account not mutable")

    code = 6064
    name = "AccountNotMutable"
    msg = "Account not mutable"


class AccountDiscriminatorMismatch(ProgramError):
    def __init__(self) -> None:
        super().__init__(6065, "Account discriminator mismatch")

    code = 6065
    name = "AccountDiscriminatorMismatch"
    msg = "Account discriminator mismatch"


class InvalidMarketNodeIndex(ProgramError):
    def __init__(self) -> None:
        super().__init__(6066, "Invalid market node index")

    code = 6066
    name = "InvalidMarketNodeIndex"
    msg = "Invalid market node index"


class InvalidMarketNode(ProgramError):
    def __init__(self) -> None:
        super().__init__(6067, "Invalid market node")

    code = 6067
    name = "InvalidMarketNode"
    msg = "Invalid market node"


class LUTOutOfBounds(ProgramError):
    def __init__(self) -> None:
        super().__init__(6068, "Lut out of bounds")

    code = 6068
    name = "LUTOutOfBounds"
    msg = "Lut out of bounds"


class RebalanceInsuranceInvalidRemainingAccounts(ProgramError):
    def __init__(self) -> None:
        super().__init__(6069, "Rebalance insurance vault with no margin accounts")

    code = 6069
    name = "RebalanceInsuranceInvalidRemainingAccounts"
    msg = "Rebalance insurance vault with no margin accounts"


class InvalidMintDecimals(ProgramError):
    def __init__(self) -> None:
        super().__init__(6070, "Invalid mint decimals")

    code = 6070
    name = "InvalidMintDecimals"
    msg = "Invalid mint decimals"


class InvalidZetaGroupOracle(ProgramError):
    def __init__(self) -> None:
        super().__init__(6071, "Invalid oracle for this zeta group")

    code = 6071
    name = "InvalidZetaGroupOracle"
    msg = "Invalid oracle for this zeta group"


class InvalidZetaGroupDepositMint(ProgramError):
    def __init__(self) -> None:
        super().__init__(6072, "Invalid zeta group deposit mint")

    code = 6072
    name = "InvalidZetaGroupDepositMint"
    msg = "Invalid zeta group deposit mint"


class InvalidZetaGroupRebalanceMint(ProgramError):
    def __init__(self) -> None:
        super().__init__(6073, "Invalid zeta group rebalance insurance vault mint")

    code = 6073
    name = "InvalidZetaGroupRebalanceMint"
    msg = "Invalid zeta group rebalance insurance vault mint"


class InvalidDepositAmount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6074, "Invalid deposit amount")

    code = 6074
    name = "InvalidDepositAmount"
    msg = "Invalid deposit amount"


class InvalidTokenAccountOwner(ProgramError):
    def __init__(self) -> None:
        super().__init__(6075, "Invalid token account owner")

    code = 6075
    name = "InvalidTokenAccountOwner"
    msg = "Invalid token account owner"


class InvalidWithdrawAmount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6076, "Invalid withdraw amount")

    code = 6076
    name = "InvalidWithdrawAmount"
    msg = "Invalid withdraw amount"


class InvalidDepositRemainingAccounts(ProgramError):
    def __init__(self) -> None:
        super().__init__(6077, "Invalid number of remaining accounts in deposit")

    code = 6077
    name = "InvalidDepositRemainingAccounts"
    msg = "Invalid number of remaining accounts in deposit"


class InvalidPlaceOrderRemainingAccounts(ProgramError):
    def __init__(self) -> None:
        super().__init__(6078, "Invalid number of remaining accounts in place order")

    code = 6078
    name = "InvalidPlaceOrderRemainingAccounts"
    msg = "Invalid number of remaining accounts in place order"


class ClientOrderIdCannotBeZero(ProgramError):
    def __init__(self) -> None:
        super().__init__(6079, "ClientOrderIdCannotBeZero")

    code = 6079
    name = "ClientOrderIdCannotBeZero"
    msg = "ClientOrderIdCannotBeZero"


class ZetaGroupHalted(ProgramError):
    def __init__(self) -> None:
        super().__init__(6080, "Zeta group halted")

    code = 6080
    name = "ZetaGroupHalted"
    msg = "Zeta group halted"


class ZetaGroupNotHalted(ProgramError):
    def __init__(self) -> None:
        super().__init__(6081, "Zeta group not halted")

    code = 6081
    name = "ZetaGroupNotHalted"
    msg = "Zeta group not halted"


class HaltMarkPriceNotSet(ProgramError):
    def __init__(self) -> None:
        super().__init__(6082, "Halt mark price not set")

    code = 6082
    name = "HaltMarkPriceNotSet"
    msg = "Halt mark price not set"


class HaltMarketsNotCleaned(ProgramError):
    def __init__(self) -> None:
        super().__init__(6083, "Halt markets not cleaned")

    code = 6083
    name = "HaltMarketsNotCleaned"
    msg = "Halt markets not cleaned"


class HaltMarketNodesNotCleaned(ProgramError):
    def __init__(self) -> None:
        super().__init__(6084, "Halt market nodes not cleaned")

    code = 6084
    name = "HaltMarketNodesNotCleaned"
    msg = "Halt market nodes not cleaned"


class CannotExpireOptionsAfterExpirationThreshold(ProgramError):
    def __init__(self) -> None:
        super().__init__(6085, "Cannot expire options after expiration threshold")

    code = 6085
    name = "CannotExpireOptionsAfterExpirationThreshold"
    msg = "Cannot expire options after expiration threshold"


class PostOnlyInCross(ProgramError):
    def __init__(self) -> None:
        super().__init__(6086, "Post only order in cross")

    code = 6086
    name = "PostOnlyInCross"
    msg = "Post only order in cross"


class FillOrKillNotFullSize(ProgramError):
    def __init__(self) -> None:
        super().__init__(6087, "Fill or kill order was not filled for full size")

    code = 6087
    name = "FillOrKillNotFullSize"
    msg = "Fill or kill order was not filled for full size"


class InvalidOpenOrdersMapOwner(ProgramError):
    def __init__(self) -> None:
        super().__init__(6088, "Invalid open orders map owner")

    code = 6088
    name = "InvalidOpenOrdersMapOwner"
    msg = "Invalid open orders map owner"


class AccountDidNotSerialize(ProgramError):
    def __init__(self) -> None:
        super().__init__(6089, "Failed to serialize the account")

    code = 6089
    name = "AccountDidNotSerialize"
    msg = "Failed to serialize the account"


class OpenOrdersWithNonEmptyPositions(ProgramError):
    def __init__(self) -> None:
        super().__init__(6090, "Cannot close open orders account with non empty positions")

    code = 6090
    name = "OpenOrdersWithNonEmptyPositions"
    msg = "Cannot close open orders account with non empty positions"


class CannotCloseNonEmptyMarginAccount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6091, "Cannot close margin account that is not empty")

    code = 6091
    name = "CannotCloseNonEmptyMarginAccount"
    msg = "Cannot close margin account that is not empty"


class InvalidTagLength(ProgramError):
    def __init__(self) -> None:
        super().__init__(6092, "Invalid tag length")

    code = 6092
    name = "InvalidTagLength"
    msg = "Invalid tag length"


class NakedShortCallIsNotAllowed(ProgramError):
    def __init__(self) -> None:
        super().__init__(6093, "Naked short call is not allowed")

    code = 6093
    name = "NakedShortCallIsNotAllowed"
    msg = "Naked short call is not allowed"


class InvalidSpreadAccount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6094, "Invalid spread account")

    code = 6094
    name = "InvalidSpreadAccount"
    msg = "Invalid spread account"


class CannotCloseNonEmptySpreadAccount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6095, "Cannot close non empty spread account")

    code = 6095
    name = "CannotCloseNonEmptySpreadAccount"
    msg = "Cannot close non empty spread account"


class SpreadAccountSeedsMismatch(ProgramError):
    def __init__(self) -> None:
        super().__init__(6096, "Spread account seeds mismatch")

    code = 6096
    name = "SpreadAccountSeedsMismatch"
    msg = "Spread account seeds mismatch"


class SpreadAccountHasUnsettledPositions(ProgramError):
    def __init__(self) -> None:
        super().__init__(6097, "Spread account seeds mismatch")

    code = 6097
    name = "SpreadAccountHasUnsettledPositions"
    msg = "Spread account seeds mismatch"


class SpreadAccountInvalidExpirySeriesState(ProgramError):
    def __init__(self) -> None:
        super().__init__(6098, "Spread account invalid expiry series state")

    code = 6098
    name = "SpreadAccountInvalidExpirySeriesState"
    msg = "Spread account invalid expiry series state"


class InsufficientFundsToCollateralizeSpreadAccount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6099, "Insufficient funds to collateralize spread account")

    code = 6099
    name = "InsufficientFundsToCollateralizeSpreadAccount"
    msg = "Insufficient funds to collateralize spread account"


class FailedMaintenanceMarginRequirement(ProgramError):
    def __init__(self) -> None:
        super().__init__(6100, "Failed maintenance margin requirement")

    code = 6100
    name = "FailedMaintenanceMarginRequirement"
    msg = "Failed maintenance margin requirement"


class InvalidMovement(ProgramError):
    def __init__(self) -> None:
        super().__init__(6101, "Invalid movement")

    code = 6101
    name = "InvalidMovement"
    msg = "Invalid movement"


class MovementOnExpiredSeries(ProgramError):
    def __init__(self) -> None:
        super().__init__(6102, "Movement on expired series")

    code = 6102
    name = "MovementOnExpiredSeries"
    msg = "Movement on expired series"


class InvalidMovementSize(ProgramError):
    def __init__(self) -> None:
        super().__init__(6103, "Invalid movement size")

    code = 6103
    name = "InvalidMovementSize"
    msg = "Invalid movement size"


class ExceededMaxPositionMovements(ProgramError):
    def __init__(self) -> None:
        super().__init__(6104, "Exceeded max position movements")

    code = 6104
    name = "ExceededMaxPositionMovements"
    msg = "Exceeded max position movements"


class ExceededMaxSpreadAccountContracts(ProgramError):
    def __init__(self) -> None:
        super().__init__(6105, "Exceeded max spread account contracts")

    code = 6105
    name = "ExceededMaxSpreadAccountContracts"
    msg = "Exceeded max spread account contracts"


class OraclePriceIsInvalid(ProgramError):
    def __init__(self) -> None:
        super().__init__(6106, "Fetched oracle price is invalid")

    code = 6106
    name = "OraclePriceIsInvalid"
    msg = "Fetched oracle price is invalid"


class InvalidUnderlyingMint(ProgramError):
    def __init__(self) -> None:
        super().__init__(6107, "Provided underlying mint address is invalid")

    code = 6107
    name = "InvalidUnderlyingMint"
    msg = "Provided underlying mint address is invalid"


class InvalidReferrerAlias(ProgramError):
    def __init__(self) -> None:
        super().__init__(6108, "Invalid referrer alias - Invalid length")

    code = 6108
    name = "InvalidReferrerAlias"
    msg = "Invalid referrer alias - Invalid length"


class ReferrerAlreadyHasAlias(ProgramError):
    def __init__(self) -> None:
        super().__init__(6109, "Referrer already has alias")

    code = 6109
    name = "ReferrerAlreadyHasAlias"
    msg = "Referrer already has alias"


class InvalidTreasuryMovementAmount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6110, "Invalid treasury movement amount")

    code = 6110
    name = "InvalidTreasuryMovementAmount"
    msg = "Invalid treasury movement amount"


class InvalidReferralsAdminSigner(ProgramError):
    def __init__(self) -> None:
        super().__init__(6111, "Invalid referrals admin signer")

    code = 6111
    name = "InvalidReferralsAdminSigner"
    msg = "Invalid referrals admin signer"


class InvalidSetReferralsRewardsRemainingAccounts(ProgramError):
    def __init__(self) -> None:
        super().__init__(6112, "Invalid set referrals rewards remaining accounts")

    code = 6112
    name = "InvalidSetReferralsRewardsRemainingAccounts"
    msg = "Invalid set referrals rewards remaining accounts"


class SetReferralsRewardsAccountNotMutable(ProgramError):
    def __init__(self) -> None:
        super().__init__(6113, "Referrals account not mutable")

    code = 6113
    name = "SetReferralsRewardsAccountNotMutable"
    msg = "Referrals account not mutable"


class InvalidClaimReferralsRewardsAmount(ProgramError):
    def __init__(self) -> None:
        super().__init__(
            6114,
            "Invalid claim referrals rewards: not enough in refererals rewards wallet",
        )

    code = 6114
    name = "InvalidClaimReferralsRewardsAmount"
    msg = "Invalid claim referrals rewards: not enough in refererals rewards wallet"


class InvalidClaimReferralsRewardsAccount(ProgramError):
    def __init__(self) -> None:
        super().__init__(
            6115,
            "Invalid claim referrals rewards: referrals account is not a referral or referrer account",
        )

    code = 6115
    name = "InvalidClaimReferralsRewardsAccount"
    msg = "Invalid claim referrals rewards: referrals account is not a referral or referrer account"


class ReferralAccountSeedsMismatch(ProgramError):
    def __init__(self) -> None:
        super().__init__(6116, "Referral account seeds mismatch")

    code = 6116
    name = "ReferralAccountSeedsMismatch"
    msg = "Referral account seeds mismatch"


class ReferrerAccountSeedsMismatch(ProgramError):
    def __init__(self) -> None:
        super().__init__(6117, "Referrer account seeds mismatch")

    code = 6117
    name = "ReferrerAccountSeedsMismatch"
    msg = "Referrer account seeds mismatch"


class ProtectedMmMarginAccount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6118, "Market maker accounts are protected from liquidation")

    code = 6118
    name = "ProtectedMmMarginAccount"
    msg = "Market maker accounts are protected from liquidation"


class CannotWithdrawWithOpenOrders(ProgramError):
    def __init__(self) -> None:
        super().__init__(6119, "Cannot withdraw with open orders")

    code = 6119
    name = "CannotWithdrawWithOpenOrders"
    msg = "Cannot withdraw with open orders"


class FundingRateNotUpToDate(ProgramError):
    def __init__(self) -> None:
        super().__init__(6120, "Perp funding rate not up to date")

    code = 6120
    name = "FundingRateNotUpToDate"
    msg = "Perp funding rate not up to date"


class PerpSyncQueueFull(ProgramError):
    def __init__(self) -> None:
        super().__init__(6121, "Perp taker/maker sync queue is full")

    code = 6121
    name = "PerpSyncQueueFull"
    msg = "Perp taker/maker sync queue is full"


class PerpSyncQueueAccountSeedsMismatch(ProgramError):
    def __init__(self) -> None:
        super().__init__(6122, "PerpSyncQueue account seeds mismatch")

    code = 6122
    name = "PerpSyncQueueAccountSeedsMismatch"
    msg = "PerpSyncQueue account seeds mismatch"


class PerpSyncQueueEmpty(ProgramError):
    def __init__(self) -> None:
        super().__init__(6123, "Program tried to pop from an empty perpSyncQueue")

    code = 6123
    name = "PerpSyncQueueEmpty"
    msg = "Program tried to pop from an empty perpSyncQueue"


class InvalidNonPerpMarket(ProgramError):
    def __init__(self) -> None:
        super().__init__(6124, "Perp product index given in placeOrder, use placePerpOrder")

    code = 6124
    name = "InvalidNonPerpMarket"
    msg = "Perp product index given in placeOrder, use placePerpOrder"


class InvalidPerpMarket(ProgramError):
    def __init__(self) -> None:
        super().__init__(6125, "Non-perp product index given in placePerpOrder, use placeOrder")

    code = 6125
    name = "InvalidPerpMarket"
    msg = "Non-perp product index given in placePerpOrder, use placeOrder"


class CannotInitializePerpMarketNode(ProgramError):
    def __init__(self) -> None:
        super().__init__(6126, "Not allowed to initialize market node for a perp market")

    code = 6126
    name = "CannotInitializePerpMarketNode"
    msg = "Not allowed to initialize market node for a perp market"


class DeprecatedInstruction(ProgramError):
    def __init__(self) -> None:
        super().__init__(6127, "Instruction is deprecated, please use the newer version")

    code = 6127
    name = "DeprecatedInstruction"
    msg = "Instruction is deprecated, please use the newer version"


class ForceCancelExpiredTIFOrdersOnly(ProgramError):
    def __init__(self) -> None:
        super().__init__(6128, "Can only force cancel expired TIF orders")

    code = 6128
    name = "ForceCancelExpiredTIFOrdersOnly"
    msg = "Can only force cancel expired TIF orders"


class InvalidPlaceOrderAuthority(ProgramError):
    def __init__(self) -> None:
        super().__init__(6129, "Invalid place order authority")

    code = 6129
    name = "InvalidPlaceOrderAuthority"
    msg = "Invalid place order authority"


class InvalidOpenOrdersAuthority(ProgramError):
    def __init__(self) -> None:
        super().__init__(6130, "Invalid open orders authority")

    code = 6130
    name = "InvalidOpenOrdersAuthority"
    msg = "Invalid open orders authority"


class InsuranceVaultSeedsMismatch(ProgramError):
    def __init__(self) -> None:
        super().__init__(6131, "Insurance vault seeds mismatch")

    code = 6131
    name = "InsuranceVaultSeedsMismatch"
    msg = "Insurance vault seeds mismatch"


class OpenInterestLimitBreach(ProgramError):
    def __init__(self) -> None:
        super().__init__(6132, "Open interest limit breach, decrease your position")

    code = 6132
    name = "OpenInterestLimitBreach"
    msg = "Open interest limit breach, decrease your position"


class WithdrawLimitBreach(ProgramError):
    def __init__(self) -> None:
        super().__init__(6133, "Withdraw limit breach, wait to withdraw more")

    code = 6133
    name = "WithdrawLimitBreach"
    msg = "Withdraw limit breach, wait to withdraw more"


class InvalidPricingOracle(ProgramError):
    def __init__(self) -> None:
        super().__init__(6134, "Invalid oracle for this pricing account")

    code = 6134
    name = "InvalidPricingOracle"
    msg = "Invalid oracle for this pricing account"


class PricingAccountSeedsMismatch(ProgramError):
    def __init__(self) -> None:
        super().__init__(6135, "Pricing account seeds mismatch")

    code = 6135
    name = "PricingAccountSeedsMismatch"
    msg = "Pricing account seeds mismatch"


class ZetaHalted(ProgramError):
    def __init__(self) -> None:
        super().__init__(6136, "Zeta exchange is halted")

    code = 6136
    name = "ZetaHalted"
    msg = "Zeta exchange is halted"


class ZetaNotHalted(ProgramError):
    def __init__(self) -> None:
        super().__init__(6137, "Zeta exchange is not halted")

    code = 6137
    name = "ZetaNotHalted"
    msg = "Zeta exchange is not halted"


class NotFreshCrossMarginAccount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6138, "Cross margin account is not unused, close it and make a new one")

    code = 6138
    name = "NotFreshCrossMarginAccount"
    msg = "Cross margin account is not unused, close it and make a new one"


class CannotCloseNonEmptyMarginAccountManager(ProgramError):
    def __init__(self) -> None:
        super().__init__(6139, "Cannot close margin account manager that is not empty")

    code = 6139
    name = "CannotCloseNonEmptyMarginAccountManager"
    msg = "Cannot close margin account manager that is not empty"


class CannotMigrateWithOpenOrders(ProgramError):
    def __init__(self) -> None:
        super().__init__(
            6140,
            "Cannot migrate to cross margin account with open orders, close all open orders",
        )

    code = 6140
    name = "CannotMigrateWithOpenOrders"
    msg = "Cannot migrate to cross margin account with open orders, close all open orders"


class InvalidMarginAccountType(ProgramError):
    def __init__(self) -> None:
        super().__init__(
            6141,
            "Invalid margin account type - account is not MarginAccount or CrossMarginAccount",
        )

    code = 6141
    name = "InvalidMarginAccountType"
    msg = "Invalid margin account type - account is not MarginAccount or CrossMarginAccount"


class MarginAccountAssetMismatch(ProgramError):
    def __init__(self) -> None:
        super().__init__(6142, "Margin account asset mismatched with instruction argument asset")

    code = 6142
    name = "MarginAccountAssetMismatch"
    msg = "Margin account asset mismatched with instruction argument asset"


class FeatureUnavailable(ProgramError):
    def __init__(self) -> None:
        super().__init__(6143, "Feature is not available yet")

    code = 6143
    name = "FeatureUnavailable"
    msg = "Feature is not available yet"


class MarginAccountCannotLiquidateCrossMarginAccount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6144, "MarginAccount cannot liquidate CrossMarginAccount")

    code = 6144
    name = "MarginAccountCannotLiquidateCrossMarginAccount"
    msg = "MarginAccount cannot liquidate CrossMarginAccount"


class InvalidDexAccOwner(ProgramError):
    def __init__(self) -> None:
        super().__init__(6145, "Invalid owner for dex account")

    code = 6145
    name = "InvalidDexAccOwner"
    msg = "Invalid owner for dex account"


class TriggerOrderCannotBeRemoved(ProgramError):
    def __init__(self) -> None:
        super().__init__(6146, "Trigger order cannot be removed")

    code = 6146
    name = "TriggerOrderCannotBeRemoved"
    msg = "Trigger order cannot be removed"


class TriggerOrderCannotBeExecuted(ProgramError):
    def __init__(self) -> None:
        super().__init__(6147, "Trigger order cannot be executed")

    code = 6147
    name = "TriggerOrderCannotBeExecuted"
    msg = "Trigger order cannot be executed"


class TooManyTriggerOrders(ProgramError):
    def __init__(self) -> None:
        super().__init__(6148, "Too many trigger orders, close some and retry")

    code = 6148
    name = "TooManyTriggerOrders"
    msg = "Too many trigger orders, close some and retry"


class InvalidTriggerOrderRemainingAccounts(ProgramError):
    def __init__(self) -> None:
        super().__init__(6149, "Invalid trigger order remaining accounts")

    code = 6149
    name = "InvalidTriggerOrderRemainingAccounts"
    msg = "Invalid trigger order remaining accounts"


class InvalidTriggerOrderWhitelistFeesAccount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6150, "Invalid trigger order whitelist fees account")

    code = 6150
    name = "InvalidTriggerOrderWhitelistFeesAccount"
    msg = "Invalid trigger order whitelist fees account"


class MissingTriggerOrderWhitelistFeesAccount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6151, "Missing trigger order whitelist fees account")

    code = 6151
    name = "MissingTriggerOrderWhitelistFeesAccount"
    msg = "Missing trigger order whitelist fees account"


class InvalidTriggerOrderBitRange(ProgramError):
    def __init__(self) -> None:
        super().__init__(6152, "Invalid trigger order bit range")

    code = 6152
    name = "InvalidTriggerOrderBitRange"
    msg = "Invalid trigger order bit range"


class InvalidSecondaryAdmin(ProgramError):
    def __init__(self) -> None:
        super().__init__(6153, "Invalid secondary admin")

    code = 6153
    name = "InvalidSecondaryAdmin"
    msg = "Invalid secondary admin"


class OnlyOwnerCanEditTriggerOrder(ProgramError):
    def __init__(self) -> None:
        super().__init__(6154, "Only the owner can edit their own trigger order")

    code = 6154
    name = "OnlyOwnerCanEditTriggerOrder"
    msg = "Only the owner can edit their own trigger order"


class TriggerOrderNeedsTimeOrPriceAndDirection(ProgramError):
    def __init__(self) -> None:
        super().__init__(
            6155,
            "Trigger order needs either a trigger price + direction, or trigger time",
        )

    code = 6155
    name = "TriggerOrderNeedsTimeOrPriceAndDirection"
    msg = "Trigger order needs either a trigger price + direction, or trigger time"


class TriggerOrderBitOccupied(ProgramError):
    def __init__(self) -> None:
        super().__init__(6156, "Given trigger order bit is occupied, pick another")

    code = 6156
    name = "TriggerOrderBitOccupied"
    msg = "Given trigger order bit is occupied, pick another"


class InvalidLiquidatorAuthority(ProgramError):
    def __init__(self) -> None:
        super().__init__(6157, "Invalid liquidator authority")

    code = 6157
    name = "InvalidLiquidatorAuthority"
    msg = "Invalid liquidator authority"


class IOCInvalidTakerFillSize(ProgramError):
    def __init__(self) -> None:
        super().__init__(6158, "IOC size_to_use doesn't match taker fill size")

    code = 6158
    name = "IOCInvalidTakerFillSize"
    msg = "IOC size_to_use doesn't match taker fill size"


class IncorrectLotSize(ProgramError):
    def __init__(self) -> None:
        super().__init__(6159, "Incorrect lot size")

    code = 6159
    name = "IncorrectLotSize"
    msg = "Incorrect lot size"


class InvalidReferrerIDLength(ProgramError):
    def __init__(self) -> None:
        super().__init__(6160, "Invalid referrer ID length")

    code = 6160
    name = "InvalidReferrerIDLength"
    msg = "Invalid referrer ID length"


class InvalidReferrerIDOwner(ProgramError):
    def __init__(self) -> None:
        super().__init__(6161, "Invalid referrer ID owner")

    code = 6161
    name = "InvalidReferrerIDOwner"
    msg = "Invalid referrer ID owner"


class CannotReferSelf(ProgramError):
    def __init__(self) -> None:
        super().__init__(6162, "User cannot refer themselves")

    code = 6162
    name = "CannotReferSelf"
    msg = "User cannot refer themselves"


class InvalidMATypeAdminSigner(ProgramError):
    def __init__(self) -> None:
        super().__init__(6163, "Invalid margin account type admin signer")

    code = 6163
    name = "InvalidMATypeAdminSigner"
    msg = "Invalid margin account type admin signer"


class PostOnlyForMulti(ProgramError):
    def __init__(self) -> None:
        super().__init__(6164, "Post only order types only for multi orders")

    code = 6164
    name = "PostOnlyForMulti"
    msg = "Post only order types only for multi orders"


class ErrTickWide(ProgramError):
    def __init__(self) -> None:
        super().__init__(6165, "ErrTickWide")

    code = 6165
    name = "ErrTickWide"
    msg = "ErrTickWide"


class OORemainingEvents(ProgramError):
    def __init__(self) -> None:
        super().__init__(6166, "Open orders has remaining events")

    code = 6166
    name = "OORemainingEvents"
    msg = "Open orders has remaining events"


class CannotForceCancelTriggerOrder(ProgramError):
    def __init__(self) -> None:
        super().__init__(6167, "Cannot force cancel trigger order")

    code = 6167
    name = "CannotForceCancelTriggerOrder"
    msg = "Cannot force cancel trigger order"


class InvalidPricingAdmin(ProgramError):
    def __init__(self) -> None:
        super().__init__(6168, "Invalid pricing admin")

    code = 6168
    name = "InvalidPricingAdmin"
    msg = "Invalid pricing admin"


class InvalidOracleUpdate(ProgramError):
    def __init__(self) -> None:
        super().__init__(6169, "Invalid oracle update")

    code = 6169
    name = "InvalidOracleUpdate"
    msg = "Invalid oracle update"


class OrderPriceTooFarFromMarkPrice(ProgramError):
    def __init__(self) -> None:
        super().__init__(6170, "Order price too far from mark price")

    code = 6170
    name = "OrderPriceTooFarFromMarkPrice"
    msg = "Order price too far from mark price"


class AirdropCommunityAlreadySet(ProgramError):
    def __init__(self) -> None:
        super().__init__(6171, "Airdrop community already set, cannot set again")

    code = 6171
    name = "AirdropCommunityAlreadySet"
    msg = "Airdrop community already set, cannot set again"


CustomError = typing.Union[
    DepositOverflow,
    Unreachable,
    FailedInitialMarginRequirement,
    LiquidatorFailedMarginRequirement,
    CannotLiquidateOwnAccount,
    CrankInvalidRemainingAccounts,
    IncorrectTickSize,
    ZeroPrice,
    ZeroSize,
    ZeroWithdrawableBalance,
    DepositAmountExceeded,
    WithdrawalAmountExceedsWithdrawableBalance,
    AccountHasSufficientMarginPostCancels,
    OverBankrupt,
    AccountHasSufficientMargin,
    UserHasNoActiveOrders,
    InvalidExpirationInterval,
    ProductMarketsAlreadyInitialized,
    InvalidProductMarketKey,
    MarketNotLive,
    MarketPricingNotReady,
    UserHasRemainingOrdersOnExpiredMarket,
    InvalidSeriesExpiration,
    InvalidExpiredOrderCancel,
    NoMarketsToAdd,
    UserHasUnsettledPositions,
    NoMarginAccountsToSettle,
    CannotSettleUserWithActiveOrders,
    OrderbookNotEmpty,
    InvalidNumberOfAccounts,
    InvalidMarketAccounts,
    ProductStrikeUninitialized,
    PricingNotUpToDate,
    RetreatsAreStale,
    ProductDirty,
    ProductStrikesInitialized,
    StrikeInitializationNotReady,
    UnsupportedKind,
    InvalidZetaGroup,
    InvalidMarginAccount,
    InvalidGreeksAccount,
    InvalidSettlementAccount,
    InvalidCancelAuthority,
    CannotUpdatePricingAfterExpiry,
    LoadAccountDiscriminatorAlreadySet,
    AccountAlreadyInitialized,
    GreeksAccountSeedsMismatch,
    ZetaGroupAccountSeedsMismatch,
    MarginAccountSeedsMismatch,
    OpenOrdersAccountSeedsMismatch,
    MarketNodeAccountSeedsMismatch,
    UserTradingFeeWhitelistAccountSeedsMismatch,
    UserDepositWhitelistAccountSeedsMismatch,
    MarketIndexesUninitialized,
    MarketIndexesAlreadyInitialized,
    CannotGetUnsetStrike,
    CannotSetInitializedStrike,
    CannotResetUninitializedStrike,
    CrankMarginAccountNotMutable,
    InvalidAdminSigner,
    UserHasActiveOrders,
    UserForceCancelInProgress,
    FailedPriceBandCheck,
    UnsortedOpenOrdersAccounts,
    AccountNotMutable,
    AccountDiscriminatorMismatch,
    InvalidMarketNodeIndex,
    InvalidMarketNode,
    LUTOutOfBounds,
    RebalanceInsuranceInvalidRemainingAccounts,
    InvalidMintDecimals,
    InvalidZetaGroupOracle,
    InvalidZetaGroupDepositMint,
    InvalidZetaGroupRebalanceMint,
    InvalidDepositAmount,
    InvalidTokenAccountOwner,
    InvalidWithdrawAmount,
    InvalidDepositRemainingAccounts,
    InvalidPlaceOrderRemainingAccounts,
    ClientOrderIdCannotBeZero,
    ZetaGroupHalted,
    ZetaGroupNotHalted,
    HaltMarkPriceNotSet,
    HaltMarketsNotCleaned,
    HaltMarketNodesNotCleaned,
    CannotExpireOptionsAfterExpirationThreshold,
    PostOnlyInCross,
    FillOrKillNotFullSize,
    InvalidOpenOrdersMapOwner,
    AccountDidNotSerialize,
    OpenOrdersWithNonEmptyPositions,
    CannotCloseNonEmptyMarginAccount,
    InvalidTagLength,
    NakedShortCallIsNotAllowed,
    InvalidSpreadAccount,
    CannotCloseNonEmptySpreadAccount,
    SpreadAccountSeedsMismatch,
    SpreadAccountHasUnsettledPositions,
    SpreadAccountInvalidExpirySeriesState,
    InsufficientFundsToCollateralizeSpreadAccount,
    FailedMaintenanceMarginRequirement,
    InvalidMovement,
    MovementOnExpiredSeries,
    InvalidMovementSize,
    ExceededMaxPositionMovements,
    ExceededMaxSpreadAccountContracts,
    OraclePriceIsInvalid,
    InvalidUnderlyingMint,
    InvalidReferrerAlias,
    ReferrerAlreadyHasAlias,
    InvalidTreasuryMovementAmount,
    InvalidReferralsAdminSigner,
    InvalidSetReferralsRewardsRemainingAccounts,
    SetReferralsRewardsAccountNotMutable,
    InvalidClaimReferralsRewardsAmount,
    InvalidClaimReferralsRewardsAccount,
    ReferralAccountSeedsMismatch,
    ReferrerAccountSeedsMismatch,
    ProtectedMmMarginAccount,
    CannotWithdrawWithOpenOrders,
    FundingRateNotUpToDate,
    PerpSyncQueueFull,
    PerpSyncQueueAccountSeedsMismatch,
    PerpSyncQueueEmpty,
    InvalidNonPerpMarket,
    InvalidPerpMarket,
    CannotInitializePerpMarketNode,
    DeprecatedInstruction,
    ForceCancelExpiredTIFOrdersOnly,
    InvalidPlaceOrderAuthority,
    InvalidOpenOrdersAuthority,
    InsuranceVaultSeedsMismatch,
    OpenInterestLimitBreach,
    WithdrawLimitBreach,
    InvalidPricingOracle,
    PricingAccountSeedsMismatch,
    ZetaHalted,
    ZetaNotHalted,
    NotFreshCrossMarginAccount,
    CannotCloseNonEmptyMarginAccountManager,
    CannotMigrateWithOpenOrders,
    InvalidMarginAccountType,
    MarginAccountAssetMismatch,
    FeatureUnavailable,
    MarginAccountCannotLiquidateCrossMarginAccount,
    InvalidDexAccOwner,
    TriggerOrderCannotBeRemoved,
    TriggerOrderCannotBeExecuted,
    TooManyTriggerOrders,
    InvalidTriggerOrderRemainingAccounts,
    InvalidTriggerOrderWhitelistFeesAccount,
    MissingTriggerOrderWhitelistFeesAccount,
    InvalidTriggerOrderBitRange,
    InvalidSecondaryAdmin,
    OnlyOwnerCanEditTriggerOrder,
    TriggerOrderNeedsTimeOrPriceAndDirection,
    TriggerOrderBitOccupied,
    InvalidLiquidatorAuthority,
    IOCInvalidTakerFillSize,
    IncorrectLotSize,
    InvalidReferrerIDLength,
    InvalidReferrerIDOwner,
    CannotReferSelf,
    InvalidMATypeAdminSigner,
    PostOnlyForMulti,
    ErrTickWide,
    OORemainingEvents,
    CannotForceCancelTriggerOrder,
    InvalidPricingAdmin,
    InvalidOracleUpdate,
    OrderPriceTooFarFromMarkPrice,
    AirdropCommunityAlreadySet,
]
CUSTOM_ERROR_MAP: dict[int, CustomError] = {
    6000: DepositOverflow(),
    6001: Unreachable(),
    6002: FailedInitialMarginRequirement(),
    6003: LiquidatorFailedMarginRequirement(),
    6004: CannotLiquidateOwnAccount(),
    6005: CrankInvalidRemainingAccounts(),
    6006: IncorrectTickSize(),
    6007: ZeroPrice(),
    6008: ZeroSize(),
    6009: ZeroWithdrawableBalance(),
    6010: DepositAmountExceeded(),
    6011: WithdrawalAmountExceedsWithdrawableBalance(),
    6012: AccountHasSufficientMarginPostCancels(),
    6013: OverBankrupt(),
    6014: AccountHasSufficientMargin(),
    6015: UserHasNoActiveOrders(),
    6016: InvalidExpirationInterval(),
    6017: ProductMarketsAlreadyInitialized(),
    6018: InvalidProductMarketKey(),
    6019: MarketNotLive(),
    6020: MarketPricingNotReady(),
    6021: UserHasRemainingOrdersOnExpiredMarket(),
    6022: InvalidSeriesExpiration(),
    6023: InvalidExpiredOrderCancel(),
    6024: NoMarketsToAdd(),
    6025: UserHasUnsettledPositions(),
    6026: NoMarginAccountsToSettle(),
    6027: CannotSettleUserWithActiveOrders(),
    6028: OrderbookNotEmpty(),
    6029: InvalidNumberOfAccounts(),
    6030: InvalidMarketAccounts(),
    6031: ProductStrikeUninitialized(),
    6032: PricingNotUpToDate(),
    6033: RetreatsAreStale(),
    6034: ProductDirty(),
    6035: ProductStrikesInitialized(),
    6036: StrikeInitializationNotReady(),
    6037: UnsupportedKind(),
    6038: InvalidZetaGroup(),
    6039: InvalidMarginAccount(),
    6040: InvalidGreeksAccount(),
    6041: InvalidSettlementAccount(),
    6042: InvalidCancelAuthority(),
    6043: CannotUpdatePricingAfterExpiry(),
    6044: LoadAccountDiscriminatorAlreadySet(),
    6045: AccountAlreadyInitialized(),
    6046: GreeksAccountSeedsMismatch(),
    6047: ZetaGroupAccountSeedsMismatch(),
    6048: MarginAccountSeedsMismatch(),
    6049: OpenOrdersAccountSeedsMismatch(),
    6050: MarketNodeAccountSeedsMismatch(),
    6051: UserTradingFeeWhitelistAccountSeedsMismatch(),
    6052: UserDepositWhitelistAccountSeedsMismatch(),
    6053: MarketIndexesUninitialized(),
    6054: MarketIndexesAlreadyInitialized(),
    6055: CannotGetUnsetStrike(),
    6056: CannotSetInitializedStrike(),
    6057: CannotResetUninitializedStrike(),
    6058: CrankMarginAccountNotMutable(),
    6059: InvalidAdminSigner(),
    6060: UserHasActiveOrders(),
    6061: UserForceCancelInProgress(),
    6062: FailedPriceBandCheck(),
    6063: UnsortedOpenOrdersAccounts(),
    6064: AccountNotMutable(),
    6065: AccountDiscriminatorMismatch(),
    6066: InvalidMarketNodeIndex(),
    6067: InvalidMarketNode(),
    6068: LUTOutOfBounds(),
    6069: RebalanceInsuranceInvalidRemainingAccounts(),
    6070: InvalidMintDecimals(),
    6071: InvalidZetaGroupOracle(),
    6072: InvalidZetaGroupDepositMint(),
    6073: InvalidZetaGroupRebalanceMint(),
    6074: InvalidDepositAmount(),
    6075: InvalidTokenAccountOwner(),
    6076: InvalidWithdrawAmount(),
    6077: InvalidDepositRemainingAccounts(),
    6078: InvalidPlaceOrderRemainingAccounts(),
    6079: ClientOrderIdCannotBeZero(),
    6080: ZetaGroupHalted(),
    6081: ZetaGroupNotHalted(),
    6082: HaltMarkPriceNotSet(),
    6083: HaltMarketsNotCleaned(),
    6084: HaltMarketNodesNotCleaned(),
    6085: CannotExpireOptionsAfterExpirationThreshold(),
    6086: PostOnlyInCross(),
    6087: FillOrKillNotFullSize(),
    6088: InvalidOpenOrdersMapOwner(),
    6089: AccountDidNotSerialize(),
    6090: OpenOrdersWithNonEmptyPositions(),
    6091: CannotCloseNonEmptyMarginAccount(),
    6092: InvalidTagLength(),
    6093: NakedShortCallIsNotAllowed(),
    6094: InvalidSpreadAccount(),
    6095: CannotCloseNonEmptySpreadAccount(),
    6096: SpreadAccountSeedsMismatch(),
    6097: SpreadAccountHasUnsettledPositions(),
    6098: SpreadAccountInvalidExpirySeriesState(),
    6099: InsufficientFundsToCollateralizeSpreadAccount(),
    6100: FailedMaintenanceMarginRequirement(),
    6101: InvalidMovement(),
    6102: MovementOnExpiredSeries(),
    6103: InvalidMovementSize(),
    6104: ExceededMaxPositionMovements(),
    6105: ExceededMaxSpreadAccountContracts(),
    6106: OraclePriceIsInvalid(),
    6107: InvalidUnderlyingMint(),
    6108: InvalidReferrerAlias(),
    6109: ReferrerAlreadyHasAlias(),
    6110: InvalidTreasuryMovementAmount(),
    6111: InvalidReferralsAdminSigner(),
    6112: InvalidSetReferralsRewardsRemainingAccounts(),
    6113: SetReferralsRewardsAccountNotMutable(),
    6114: InvalidClaimReferralsRewardsAmount(),
    6115: InvalidClaimReferralsRewardsAccount(),
    6116: ReferralAccountSeedsMismatch(),
    6117: ReferrerAccountSeedsMismatch(),
    6118: ProtectedMmMarginAccount(),
    6119: CannotWithdrawWithOpenOrders(),
    6120: FundingRateNotUpToDate(),
    6121: PerpSyncQueueFull(),
    6122: PerpSyncQueueAccountSeedsMismatch(),
    6123: PerpSyncQueueEmpty(),
    6124: InvalidNonPerpMarket(),
    6125: InvalidPerpMarket(),
    6126: CannotInitializePerpMarketNode(),
    6127: DeprecatedInstruction(),
    6128: ForceCancelExpiredTIFOrdersOnly(),
    6129: InvalidPlaceOrderAuthority(),
    6130: InvalidOpenOrdersAuthority(),
    6131: InsuranceVaultSeedsMismatch(),
    6132: OpenInterestLimitBreach(),
    6133: WithdrawLimitBreach(),
    6134: InvalidPricingOracle(),
    6135: PricingAccountSeedsMismatch(),
    6136: ZetaHalted(),
    6137: ZetaNotHalted(),
    6138: NotFreshCrossMarginAccount(),
    6139: CannotCloseNonEmptyMarginAccountManager(),
    6140: CannotMigrateWithOpenOrders(),
    6141: InvalidMarginAccountType(),
    6142: MarginAccountAssetMismatch(),
    6143: FeatureUnavailable(),
    6144: MarginAccountCannotLiquidateCrossMarginAccount(),
    6145: InvalidDexAccOwner(),
    6146: TriggerOrderCannotBeRemoved(),
    6147: TriggerOrderCannotBeExecuted(),
    6148: TooManyTriggerOrders(),
    6149: InvalidTriggerOrderRemainingAccounts(),
    6150: InvalidTriggerOrderWhitelistFeesAccount(),
    6151: MissingTriggerOrderWhitelistFeesAccount(),
    6152: InvalidTriggerOrderBitRange(),
    6153: InvalidSecondaryAdmin(),
    6154: OnlyOwnerCanEditTriggerOrder(),
    6155: TriggerOrderNeedsTimeOrPriceAndDirection(),
    6156: TriggerOrderBitOccupied(),
    6157: InvalidLiquidatorAuthority(),
    6158: IOCInvalidTakerFillSize(),
    6159: IncorrectLotSize(),
    6160: InvalidReferrerIDLength(),
    6161: InvalidReferrerIDOwner(),
    6162: CannotReferSelf(),
    6163: InvalidMATypeAdminSigner(),
    6164: PostOnlyForMulti(),
    6165: ErrTickWide(),
    6166: OORemainingEvents(),
    6167: CannotForceCancelTriggerOrder(),
    6168: InvalidPricingAdmin(),
    6169: InvalidOracleUpdate(),
    6170: OrderPriceTooFarFromMarkPrice(),
    6171: AirdropCommunityAlreadySet(),
}


def from_code(code: int) -> typing.Optional[CustomError]:
    maybe_err = CUSTOM_ERROR_MAP.get(code)
    if maybe_err is None:
        return None
    return maybe_err



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container

from . import asset


class InitializeZetaGroupArgsJSON(typing.TypedDict):
    perps_only: bool
    flex_underlying: bool
    asset_override: typing.Optional[asset.AssetJSON]
    zeta_group_nonce: int
    underlying_nonce: int
    greeks_nonce: int
    vault_nonce: int
    insurance_vault_nonce: int
    socialized_loss_account_nonce: int
    perp_sync_queue_nonce: int
    interest_rate: int
    volatility: list[int]
    option_trade_normalizer: int
    future_trade_normalizer: int
    max_volatility_retreat: int
    max_interest_retreat: int
    max_delta: int
    min_delta: int
    min_interest_rate: int
    max_interest_rate: int
    min_volatility: int
    max_volatility: int
    future_margin_initial: int
    future_margin_maintenance: int
    expiry_interval_seconds: int
    new_expiry_threshold_seconds: int
    min_funding_rate_percent: int
    max_funding_rate_percent: int
    perp_impact_cash_delta: int


@dataclass
class InitializeZetaGroupArgs:
    layout: typing.ClassVar = borsh.CStruct(
        "perps_only" / borsh.Bool,
        "flex_underlying" / borsh.Bool,
        "asset_override" / borsh.Option(asset.layout),
        "zeta_group_nonce" / borsh.U8,
        "underlying_nonce" / borsh.U8,
        "greeks_nonce" / borsh.U8,
        "vault_nonce" / borsh.U8,
        "insurance_vault_nonce" / borsh.U8,
        "socialized_loss_account_nonce" / borsh.U8,
        "perp_sync_queue_nonce" / borsh.U8,
        "interest_rate" / borsh.I64,
        "volatility" / borsh.U64[5],
        "option_trade_normalizer" / borsh.U64,
        "future_trade_normalizer" / borsh.U64,
        "max_volatility_retreat" / borsh.U64,
        "max_interest_retreat" / borsh.U64,
        "max_delta" / borsh.U64,
        "min_delta" / borsh.U64,
        "min_interest_rate" / borsh.I64,
        "max_interest_rate" / borsh.I64,
        "min_volatility" / borsh.U64,
        "max_volatility" / borsh.U64,
        "future_margin_initial" / borsh.U64,
        "future_margin_maintenance" / borsh.U64,
        "expiry_interval_seconds" / borsh.U32,
        "new_expiry_threshold_seconds" / borsh.U32,
        "min_funding_rate_percent" / borsh.I64,
        "max_funding_rate_percent" / borsh.I64,
        "perp_impact_cash_delta" / borsh.U64,
    )
    perps_only: bool
    flex_underlying: bool
    asset_override: typing.Optional[asset.AssetKind]
    zeta_group_nonce: int
    underlying_nonce: int
    greeks_nonce: int
    vault_nonce: int
    insurance_vault_nonce: int
    socialized_loss_account_nonce: int
    perp_sync_queue_nonce: int
    interest_rate: int
    volatility: list[int]
    option_trade_normalizer: int
    future_trade_normalizer: int
    max_volatility_retreat: int
    max_interest_retreat: int
    max_delta: int
    min_delta: int
    min_interest_rate: int
    max_interest_rate: int
    min_volatility: int
    max_volatility: int
    future_margin_initial: int
    future_margin_maintenance: int
    expiry_interval_seconds: int
    new_expiry_threshold_seconds: int
    min_funding_rate_percent: int
    max_funding_rate_percent: int
    perp_impact_cash_delta: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "InitializeZetaGroupArgs":
        return cls(
            perps_only=obj.perps_only,
            flex_underlying=obj.flex_underlying,
            asset_override=(None if obj.asset_override is None else asset.from_decoded(obj.asset_override)),
            zeta_group_nonce=obj.zeta_group_nonce,
            underlying_nonce=obj.underlying_nonce,
            greeks_nonce=obj.greeks_nonce,
            vault_nonce=obj.vault_nonce,
            insurance_vault_nonce=obj.insurance_vault_nonce,
            socialized_loss_account_nonce=obj.socialized_loss_account_nonce,
            perp_sync_queue_nonce=obj.perp_sync_queue_nonce,
            interest_rate=obj.interest_rate,
            volatility=obj.volatility,
            option_trade_normalizer=obj.option_trade_normalizer,
            future_trade_normalizer=obj.future_trade_normalizer,
            max_volatility_retreat=obj.max_volatility_retreat,
            max_interest_retreat=obj.max_interest_retreat,
            max_delta=obj.max_delta,
            min_delta=obj.min_delta,
            min_interest_rate=obj.min_interest_rate,
            max_interest_rate=obj.max_interest_rate,
            min_volatility=obj.min_volatility,
            max_volatility=obj.max_volatility,
            future_margin_initial=obj.future_margin_initial,
            future_margin_maintenance=obj.future_margin_maintenance,
            expiry_interval_seconds=obj.expiry_interval_seconds,
            new_expiry_threshold_seconds=obj.new_expiry_threshold_seconds,
            min_funding_rate_percent=obj.min_funding_rate_percent,
            max_funding_rate_percent=obj.max_funding_rate_percent,
            perp_impact_cash_delta=obj.perp_impact_cash_delta,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "perps_only": self.perps_only,
            "flex_underlying": self.flex_underlying,
            "asset_override": (None if self.asset_override is None else self.asset_override.to_encodable()),
            "zeta_group_nonce": self.zeta_group_nonce,
            "underlying_nonce": self.underlying_nonce,
            "greeks_nonce": self.greeks_nonce,
            "vault_nonce": self.vault_nonce,
            "insurance_vault_nonce": self.insurance_vault_nonce,
            "socialized_loss_account_nonce": self.socialized_loss_account_nonce,
            "perp_sync_queue_nonce": self.perp_sync_queue_nonce,
            "interest_rate": self.interest_rate,
            "volatility": self.volatility,
            "option_trade_normalizer": self.option_trade_normalizer,
            "future_trade_normalizer": self.future_trade_normalizer,
            "max_volatility_retreat": self.max_volatility_retreat,
            "max_interest_retreat": self.max_interest_retreat,
            "max_delta": self.max_delta,
            "min_delta": self.min_delta,
            "min_interest_rate": self.min_interest_rate,
            "max_interest_rate": self.max_interest_rate,
            "min_volatility": self.min_volatility,
            "max_volatility": self.max_volatility,
            "future_margin_initial": self.future_margin_initial,
            "future_margin_maintenance": self.future_margin_maintenance,
            "expiry_interval_seconds": self.expiry_interval_seconds,
            "new_expiry_threshold_seconds": self.new_expiry_threshold_seconds,
            "min_funding_rate_percent": self.min_funding_rate_percent,
            "max_funding_rate_percent": self.max_funding_rate_percent,
            "perp_impact_cash_delta": self.perp_impact_cash_delta,
        }

    def to_json(self) -> InitializeZetaGroupArgsJSON:
        return {
            "perps_only": self.perps_only,
            "flex_underlying": self.flex_underlying,
            "asset_override": (None if self.asset_override is None else self.asset_override.to_json()),
            "zeta_group_nonce": self.zeta_group_nonce,
            "underlying_nonce": self.underlying_nonce,
            "greeks_nonce": self.greeks_nonce,
            "vault_nonce": self.vault_nonce,
            "insurance_vault_nonce": self.insurance_vault_nonce,
            "socialized_loss_account_nonce": self.socialized_loss_account_nonce,
            "perp_sync_queue_nonce": self.perp_sync_queue_nonce,
            "interest_rate": self.interest_rate,
            "volatility": self.volatility,
            "option_trade_normalizer": self.option_trade_normalizer,
            "future_trade_normalizer": self.future_trade_normalizer,
            "max_volatility_retreat": self.max_volatility_retreat,
            "max_interest_retreat": self.max_interest_retreat,
            "max_delta": self.max_delta,
            "min_delta": self.min_delta,
            "min_interest_rate": self.min_interest_rate,
            "max_interest_rate": self.max_interest_rate,
            "min_volatility": self.min_volatility,
            "max_volatility": self.max_volatility,
            "future_margin_initial": self.future_margin_initial,
            "future_margin_maintenance": self.future_margin_maintenance,
            "expiry_interval_seconds": self.expiry_interval_seconds,
            "new_expiry_threshold_seconds": self.new_expiry_threshold_seconds,
            "min_funding_rate_percent": self.min_funding_rate_percent,
            "max_funding_rate_percent": self.max_funding_rate_percent,
            "perp_impact_cash_delta": self.perp_impact_cash_delta,
        }

    @classmethod
    def from_json(cls, obj: InitializeZetaGroupArgsJSON) -> "InitializeZetaGroupArgs":
        return cls(
            perps_only=obj["perps_only"],
            flex_underlying=obj["flex_underlying"],
            asset_override=(None if obj["asset_override"] is None else asset.from_json(obj["asset_override"])),
            zeta_group_nonce=obj["zeta_group_nonce"],
            underlying_nonce=obj["underlying_nonce"],
            greeks_nonce=obj["greeks_nonce"],
            vault_nonce=obj["vault_nonce"],
            insurance_vault_nonce=obj["insurance_vault_nonce"],
            socialized_loss_account_nonce=obj["socialized_loss_account_nonce"],
            perp_sync_queue_nonce=obj["perp_sync_queue_nonce"],
            interest_rate=obj["interest_rate"],
            volatility=obj["volatility"],
            option_trade_normalizer=obj["option_trade_normalizer"],
            future_trade_normalizer=obj["future_trade_normalizer"],
            max_volatility_retreat=obj["max_volatility_retreat"],
            max_interest_retreat=obj["max_interest_retreat"],
            max_delta=obj["max_delta"],
            min_delta=obj["min_delta"],
            min_interest_rate=obj["min_interest_rate"],
            max_interest_rate=obj["max_interest_rate"],
            min_volatility=obj["min_volatility"],
            max_volatility=obj["max_volatility"],
            future_margin_initial=obj["future_margin_initial"],
            future_margin_maintenance=obj["future_margin_maintenance"],
            expiry_interval_seconds=obj["expiry_interval_seconds"],
            new_expiry_threshold_seconds=obj["new_expiry_threshold_seconds"],
            min_funding_rate_percent=obj["min_funding_rate_percent"],
            max_funding_rate_percent=obj["max_funding_rate_percent"],
            perp_impact_cash_delta=obj["perp_impact_cash_delta"],
        )



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container


class OrderArgsJSON(typing.TypedDict):
    price: int
    size: int
    client_order_id: typing.Optional[int]
    tif_offset: typing.Optional[int]


@dataclass
class OrderArgs:
    layout: typing.ClassVar = borsh.CStruct(
        "price" / borsh.U64,
        "size" / borsh.U64,
        "client_order_id" / borsh.Option(borsh.U64),
        "tif_offset" / borsh.Option(borsh.U16),
    )
    price: int
    size: int
    client_order_id: typing.Optional[int]
    tif_offset: typing.Optional[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "OrderArgs":
        return cls(
            price=obj.price,
            size=obj.size,
            client_order_id=obj.client_order_id,
            tif_offset=obj.tif_offset,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "price": self.price,
            "size": self.size,
            "client_order_id": self.client_order_id,
            "tif_offset": self.tif_offset,
        }

    def to_json(self) -> OrderArgsJSON:
        return {
            "price": self.price,
            "size": self.size,
            "client_order_id": self.client_order_id,
            "tif_offset": self.tif_offset,
        }

    @classmethod
    def from_json(cls, obj: OrderArgsJSON) -> "OrderArgs":
        return cls(
            price=obj["price"],
            size=obj["size"],
            client_order_id=obj["client_order_id"],
            tif_offset=obj["tif_offset"],
        )



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from solders.sysvar import RENT
from spl.token.constants import TOKEN_PROGRAM_ID

from .. import types
from ..program_id import PROGRAM_ID


class InitializeZetaGroupArgs(typing.TypedDict):
    args: types.initialize_zeta_group_args.InitializeZetaGroupArgs


layout = borsh.CStruct("args" / types.initialize_zeta_group_args.InitializeZetaGroupArgs.layout)


class InitializeZetaGroupAccounts(typing.TypedDict):
    state: Pubkey
    admin: Pubkey
    underlying_mint: Pubkey
    zeta_program: Pubkey
    oracle: Pubkey
    oracle_backup_feed: Pubkey
    oracle_backup_program: Pubkey
    zeta_group: Pubkey
    greeks: Pubkey
    perp_sync_queue: Pubkey
    underlying: Pubkey
    vault: Pubkey
    insurance_vault: Pubkey
    socialized_loss_account: Pubkey
    usdc_mint: Pubkey


def initialize_zeta_group(
    args: InitializeZetaGroupArgs,
    accounts: InitializeZetaGroupAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["underlying_mint"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["zeta_program"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["oracle"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["oracle_backup_feed"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["oracle_backup_program"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["zeta_group"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["greeks"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["perp_sync_queue"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["underlying"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["vault"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["insurance_vault"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["socialized_loss_account"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["usdc_mint"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=RENT, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x06\x87$\xe8#'\xfaG"
    encoded_args = layout.build(
        {
            "args": args["args"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import EnumForCodegen


class UninitializedJSON(typing.TypedDict):
    kind: typing.Literal["Uninitialized"]


class CallJSON(typing.TypedDict):
    kind: typing.Literal["Call"]


class PutJSON(typing.TypedDict):
    kind: typing.Literal["Put"]


class FutureJSON(typing.TypedDict):
    kind: typing.Literal["Future"]


class PerpJSON(typing.TypedDict):
    kind: typing.Literal["Perp"]


@dataclass
class Uninitialized:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Uninitialized"

    @classmethod
    def to_json(cls) -> UninitializedJSON:
        return UninitializedJSON(
            kind="Uninitialized",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Uninitialized": {},
        }


@dataclass
class Call:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "Call"

    @classmethod
    def to_json(cls) -> CallJSON:
        return CallJSON(
            kind="Call",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Call": {},
        }


@dataclass
class Put:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "Put"

    @classmethod
    def to_json(cls) -> PutJSON:
        return PutJSON(
            kind="Put",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Put": {},
        }


@dataclass
class Future:
    discriminator: typing.ClassVar = 3
    kind: typing.ClassVar = "Future"

    @classmethod
    def to_json(cls) -> FutureJSON:
        return FutureJSON(
            kind="Future",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Future": {},
        }


@dataclass
class Perp:
    discriminator: typing.ClassVar = 4
    kind: typing.ClassVar = "Perp"

    @classmethod
    def to_json(cls) -> PerpJSON:
        return PerpJSON(
            kind="Perp",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Perp": {},
        }


KindKind = typing.Union[Uninitialized, Call, Put, Future, Perp]
KindJSON = typing.Union[UninitializedJSON, CallJSON, PutJSON, FutureJSON, PerpJSON]


def from_decoded(obj: dict) -> KindKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Uninitialized" in obj:
        return Uninitialized()
    if "Call" in obj:
        return Call()
    if "Put" in obj:
        return Put()
    if "Future" in obj:
        return Future()
    if "Perp" in obj:
        return Perp()
    raise ValueError("Invalid enum object")


def from_json(obj: KindJSON) -> KindKind:
    if obj["kind"] == "Uninitialized":
        return Uninitialized()
    if obj["kind"] == "Call":
        return Call()
    if obj["kind"] == "Put":
        return Put()
    if obj["kind"] == "Future":
        return Future()
    if obj["kind"] == "Perp":
        return Perp()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "Uninitialized" / borsh.CStruct(),
    "Call" / borsh.CStruct(),
    "Put" / borsh.CStruct(),
    "Future" / borsh.CStruct(),
    "Perp" / borsh.CStruct(),
)



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container


class HaltStateV2JSON(typing.TypedDict):
    halted: bool
    timestamp: int
    spot_price: int
    market_cleaned: bool


@dataclass
class HaltStateV2:
    layout: typing.ClassVar = borsh.CStruct(
        "halted" / borsh.Bool,
        "timestamp" / borsh.U64,
        "spot_price" / borsh.U64,
        "market_cleaned" / borsh.Bool,
    )
    halted: bool
    timestamp: int
    spot_price: int
    market_cleaned: bool

    @classmethod
    def from_decoded(cls, obj: Container) -> "HaltStateV2":
        return cls(
            halted=obj.halted,
            timestamp=obj.timestamp,
            spot_price=obj.spot_price,
            market_cleaned=obj.market_cleaned,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "halted": self.halted,
            "timestamp": self.timestamp,
            "spot_price": self.spot_price,
            "market_cleaned": self.market_cleaned,
        }

    def to_json(self) -> HaltStateV2JSON:
        return {
            "halted": self.halted,
            "timestamp": self.timestamp,
            "spot_price": self.spot_price,
            "market_cleaned": self.market_cleaned,
        }

    @classmethod
    def from_json(cls, obj: HaltStateV2JSON) -> "HaltStateV2":
        return cls(
            halted=obj["halted"],
            timestamp=obj["timestamp"],
            spot_price=obj["spot_price"],
            market_cleaned=obj["market_cleaned"],
        )



from dataclasses import dataclass

from zetamarkets_py import constants, utils
from zetamarkets_py.types import Asset
from zetamarkets_py.zeta_client.accounts.cross_margin_account import CrossMarginAccount
from zetamarkets_py.zeta_client.accounts.pricing import Pricing


def compute_upnl(margin_account: CrossMarginAccount, pricing_account: Pricing) -> float:
    """
    Compute the unrealized profit and loss (upnl) of the account.

    Args:
        margin_account (CrossMarginAccount): The margin account of the account.
        pricing_account (Pricing): The pricing account of the account.

    Returns:
        float: The upnl of the account.
    """
    positions = [Position.from_margin_account(margin_account, i) for i in range(len(margin_account.product_ledgers))]
    mark_prices = [utils.convert_fixed_int_to_decimal(p) for p in pricing_account.mark_prices]

    upnl = sum(
        [
            (
                (pos.size * mark_price - pos.cost_of_trades)
                if pos.size > 0
                else (pos.size * mark_price + pos.cost_of_trades)
            )
            for pos, mark_price in zip(positions, mark_prices)
        ]
    )
    return upnl


def compute_margin_usage(margin_account: CrossMarginAccount, pricing_account: Pricing) -> float:
    """
    Compute the margin usage of the account.

    Args:
        positions (dict[Asset, Position]): The positions of the account.
        mark_prices (dict[Asset, float]): The mark prices of the assets.

    Returns:
        float: The margin usage of the account.
    """
    positions = [Position.from_margin_account(margin_account, i) for i in range(len(margin_account.product_ledgers))]
    mark_prices = [utils.convert_fixed_int_to_decimal(p) for p in pricing_account.mark_prices]

    return sum([abs(pos.size) * mark_price for pos, mark_price in zip(positions, mark_prices)])


def compute_initial_margin(margin_account: CrossMarginAccount, pricing_account: Pricing) -> float:
    """
    Compute the initial margin of the account.

    Args:
        margin_account (CrossMarginAccount): The margin account of the account.
        mark_prices (dict): The mark prices of the assets.
        margin_params (dict): The margin parameters of the assets.

    Returns:
        float: The initial margin of the account.
    """
    initial_margin = 0
    for ledger, mark_price_raw, params in zip(
        margin_account.product_ledgers, pricing_account.mark_prices, pricing_account.margin_parameters
    ):
        size = utils.convert_fixed_lot_to_decimal(ledger.position.size)
        mark_price = utils.convert_fixed_int_to_decimal(mark_price_raw)
        long_lots = utils.convert_fixed_lot_to_decimal(ledger.order_state.opening_orders[0])
        short_lots = utils.convert_fixed_lot_to_decimal(ledger.order_state.opening_orders[1])
        initial_margin_ratio = params.future_margin_initial / 10**constants.MARGIN_PRECISION

        if size == 0 and long_lots == 0 and short_lots == 0:
            continue

        if size > 0:
            long_lots += abs(size)
        elif size < 0:
            short_lots += abs(size)

        lots = long_lots if long_lots > short_lots else short_lots
        initial_margin += abs(lots) * mark_price * initial_margin_ratio
    return initial_margin


def compute_maintenance_margin(margin_account: CrossMarginAccount, pricing_account: Pricing) -> float:
    """
    Compute the maintenance margin of the account.

    Args:
        positions (dict): The positions of the account.
        mark_prices (dict): The mark prices of the assets.
        margin_params (dict): The margin parameters of the assets.

    Returns:
        float: The maintenance margin of the account.
    """
    positions = [Position.from_margin_account(margin_account, i) for i in range(len(margin_account.product_ledgers))]
    mark_prices = [utils.convert_fixed_int_to_decimal(p) for p in pricing_account.mark_prices]
    maintenance_margin_params = [
        params.future_margin_maintenance / 10**constants.MARGIN_PRECISION
        for params in pricing_account.margin_parameters
    ]
    return sum(
        [
            abs(pos.size) * mark_price * maintenance_margin_ratio
            for pos, mark_price, maintenance_margin_ratio in zip(positions, mark_prices, maintenance_margin_params)
        ]
    )


@dataclass
class Position:
    """Data class for position details."""

    size: float
    cost_of_trades: float

    @property
    def average_price(self) -> float:
        return self.cost_of_trades / self.size

    @classmethod
    def from_margin_account(cls, margin_account: CrossMarginAccount, asset_index: int):
        return cls(
            utils.convert_fixed_lot_to_decimal(margin_account.product_ledgers[asset_index].position.size),
            utils.convert_fixed_int_to_decimal(margin_account.product_ledgers[asset_index].position.cost_of_trades),
        )


@dataclass
class AccountRiskSummary:
    balance: float
    unrealized_pnl: float
    equity: float
    position_value: float
    initial_margin: float
    maintenance_margin: float
    margin_utilization: float
    leverage: float
    positions: dict[Asset, Position]
    mark_prices: dict[Asset, float]

    @classmethod
    def from_margin_and_pricing_accounts(cls, margin_account: CrossMarginAccount, pricing_account: Pricing):
        balance = utils.convert_fixed_int_to_decimal(margin_account.balance)
        upnl = compute_upnl(margin_account, pricing_account)
        equity = balance + upnl
        position_value = compute_margin_usage(margin_account, pricing_account)
        initial_margin = compute_initial_margin(margin_account, pricing_account)
        maintenance_margin = compute_maintenance_margin(margin_account, pricing_account)
        margin_utilization = maintenance_margin / equity
        leverage = position_value / equity

        # TODO: decide on list vs dict representation
        positions = {
            Asset.from_index(i): Position.from_margin_account(margin_account, i)
            for i in range(len(margin_account.product_ledgers))
        }
        mark_prices = {
            a: utils.convert_fixed_int_to_decimal(p) for a, p in zip(Asset.all(), pricing_account.mark_prices)
        }

        return AccountRiskSummary(
            balance,
            upnl,
            equity,
            position_value,
            initial_margin,
            maintenance_margin,
            margin_utilization,
            leverage,
            positions,
            mark_prices,
        )



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container


class OrderStateJSON(typing.TypedDict):
    closing_orders: int
    opening_orders: list[int]


@dataclass
class OrderState:
    layout: typing.ClassVar = borsh.CStruct("closing_orders" / borsh.U64, "opening_orders" / borsh.U64[2])
    closing_orders: int
    opening_orders: list[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "OrderState":
        return cls(closing_orders=obj.closing_orders, opening_orders=obj.opening_orders)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "closing_orders": self.closing_orders,
            "opening_orders": self.opening_orders,
        }

    def to_json(self) -> OrderStateJSON:
        return {
            "closing_orders": self.closing_orders,
            "opening_orders": self.opening_orders,
        }

    @classmethod
    def from_json(cls, obj: OrderStateJSON) -> "OrderState":
        return cls(closing_orders=obj["closing_orders"], opening_orders=obj["opening_orders"])



from dataclasses import dataclass
from typing import Any, Optional, Union

from anchorpy import Event
from construct import Container
from solders.pubkey import Pubkey

from zetamarkets_py import utils
from zetamarkets_py.types import Asset, MultiOrderArgs, OrderCompleteType, Side


# PlaceOrderEvent but we add the args from the PlaceOrder instruction itself, as well as the tx slot and signature
@dataclass
class PlaceOrderEventWithArgs:
    """Program event for placing an order.

    Note:
        This class is an extension of the :class:`PlaceOrderEvent` class. It includes additional arguments from the
        PlaceOrder instruction.
    """

    # ix args
    price: float
    size: float
    side: Side

    # ix event
    fee: float
    oracle_price: float
    order_id: int
    asset: Asset
    margin_account: Pubkey
    client_order_id: int

    @classmethod
    def from_event_and_args(cls, event: Event, args: Container):
        assert event.name.startswith("PlaceOrderEvent")
        return cls(
            price=utils.convert_fixed_int_to_decimal(args.price),
            size=utils.convert_fixed_lot_to_decimal(args.size),
            side=Side.from_index(args.side.index),
            fee=utils.convert_fixed_int_to_decimal(event.data.fee),
            oracle_price=utils.convert_fixed_int_to_decimal(event.data.oracle_price),
            order_id=event.data.order_id,
            asset=Asset.from_index(event.data.asset.index),
            margin_account=event.data.margin_account,
            client_order_id=event.data.client_order_id,
        )


@dataclass
class PlaceOrderEvent:
    """Program event for placing an order."""

    fee: float
    oracle_price: float
    order_id: int
    asset: Asset
    margin_account: Pubkey
    client_order_id: int

    @classmethod
    def from_event(cls, event: Event):
        assert event.name == cls.__name__
        return cls(
            fee=utils.convert_fixed_int_to_decimal(event.data.fee),
            oracle_price=utils.convert_fixed_int_to_decimal(event.data.oracle_price),
            order_id=event.data.order_id,
            asset=Asset.from_index(event.data.asset.index),
            margin_account=event.data.margin_account,
            client_order_id=event.data.client_order_id,
        )


@dataclass
class PlaceMultiOrdersEventWithArgs:
    """Program event for placing multiple orders.

    Note:
        This class is an extension of the :class:`PlaceMultiOrdersEvent` class. It includes additional arguments from the
        PlaceMultiOrders instruction.
    """

    # ix args
    bid_orders: list[MultiOrderArgs]
    ask_orders: list[MultiOrderArgs]

    oracle_price: float
    order_ids: list[int]
    expiry_tss: list[int]
    asset: Asset
    margin_account: Pubkey
    client_order_ids: list[int]
    user: Pubkey

    @classmethod
    def from_event_and_args(cls, event: Event, args: Container):
        assert event.name.startswith("PlaceMultiOrdersEvent")

        for order in args.bid_orders:
            order.price = utils.convert_fixed_int_to_decimal(order.price)
            order.size = utils.convert_fixed_lot_to_decimal(order.size)
        for order in args.ask_orders:
            order.price = utils.convert_fixed_int_to_decimal(order.price)
            order.size = utils.convert_fixed_lot_to_decimal(order.size)

        return cls(
            bid_orders=args.bid_orders,
            ask_orders=args.ask_orders,
            oracle_price=utils.convert_fixed_int_to_decimal(event.data.oracle_price),
            order_ids=event.data.order_ids,
            expiry_tss=event.data.expiry_tss,
            asset=Asset.from_index(event.data.asset.index),
            margin_account=event.data.margin_account,
            client_order_ids=event.data.client_order_ids,
            user=event.data.user,
        )


@dataclass
class PlaceMultiOrdersEvent:
    """Program event for placing multiple orders."""

    oracle_price: float
    order_ids: list[int]
    expiry_tss: list[int]
    asset: Asset
    margin_account: Pubkey
    client_order_ids: list[int]
    user: Pubkey

    @classmethod
    def from_event(cls, event: Event):
        assert event.name == cls.__name__
        return cls(
            oracle_price=utils.convert_fixed_int_to_decimal(event.data.oracle_price),
            order_ids=event.data.order_ids,
            expiry_tss=event.data.expiry_tss,
            asset=Asset.from_index(event.data.asset.index),
            margin_account=event.data.margin_account,
            client_order_ids=event.data.client_order_ids,
            user=event.data.user,
        )


@dataclass
class OrderCompleteEvent:
    """Program event for an order being completed.

    Note: This event is emitted when an order is either fully filled, cancelled or booted (i.e. TIF expiry).
    """

    margin_account: Pubkey
    authority: Pubkey
    asset: Asset
    side: Side
    unfilled_size: float
    order_id: int
    client_order_id: int
    order_complete_type: OrderCompleteType

    @classmethod
    def from_event(cls, event: Event):
        assert event.name == cls.__name__
        return cls(
            margin_account=event.data.margin_account,
            authority=event.data.user,
            asset=Asset.from_index(event.data.asset.index),
            side=Side.from_index(event.data.side.index),
            unfilled_size=utils.convert_fixed_lot_to_decimal(event.data.unfilled_size),
            order_id=event.data.order_id,
            client_order_id=event.data.client_order_id,
            order_complete_type=OrderCompleteType.from_index(event.data.order_complete_type.index),
        )


@dataclass
class CancelOrderEvent:
    """Event for cancelling an order.

    Note: This event is emitted when an order is cancelled, including auto-cancels like TIF.
    """

    margin_account: Pubkey
    authority: Pubkey
    asset: Asset
    side: Side
    unfilled_size: float
    order_id: int
    client_order_id: int

    @classmethod
    def from_order_complete_event(cls, event: OrderCompleteEvent):
        return cls(
            margin_account=event.margin_account,
            authority=event.authority,
            asset=event.asset,
            side=event.side,
            unfilled_size=event.unfilled_size,
            order_id=event.order_id,
            client_order_id=event.client_order_id,
        )


@dataclass
class TradeEvent:
    """Program event for a trade."""

    margin_account: Pubkey
    price: float
    size: float
    cost_of_trades: float
    side: Side
    client_order_id: int
    order_id: int
    asset: Asset
    authority: Pubkey
    is_taker: bool
    sequence_number: int
    fee: float
    pnl: float
    rebate: float

    @classmethod
    def from_event(cls, event: Event):
        assert event.name.startswith(cls.__name__)
        return cls(
            margin_account=event.data.margin_account,
            price=utils.convert_fixed_int_to_decimal(event.data.price),
            size=utils.convert_fixed_lot_to_decimal(event.data.size),
            cost_of_trades=utils.convert_fixed_int_to_decimal(event.data.cost_of_trades),
            side=Side.Bid if event.data.is_bid else Side.Ask,
            client_order_id=event.data.client_order_id,
            order_id=event.data.order_id,
            asset=Asset.from_index(event.data.asset.index),
            authority=event.data.user,
            is_taker=event.data.is_taker,
            sequence_number=event.data.sequence_number,
            fee=utils.convert_fixed_int_to_decimal(event.data.fee),
            pnl=utils.convert_fixed_int_to_decimal(event.data.pnl),
            rebate=utils.convert_fixed_int_to_decimal(event.data.rebate),
        )


@dataclass
class LiquidationEvent:
    """Program event for a liquidation."""

    liquidator_reward: float
    insurance_reward: float
    side: Side
    liquidation_price: float
    liquidation_size: float
    remaining_liquidatee_balance: float
    remaining_liquidator_balance: float
    mark_price: float
    oracle_price: float
    liquidatee: Pubkey
    liquidator: Pubkey
    asset: Asset
    liquidatee_margin_account: Pubkey
    liquidator_margin_account: Pubkey

    @classmethod
    def from_event(cls, event: Event):
        assert event.name == cls.__name__
        return cls(
            liquidator_reward=utils.convert_fixed_int_to_decimal(event.data.liquidator_reward),
            insurance_reward=utils.convert_fixed_int_to_decimal(event.data.insurance_reward),
            side=Side.Bid if event.data.size > 0 else Side.Ask,
            liquidation_price=utils.convert_fixed_int_to_decimal(event.data.cost_of_trades)
            / utils.convert_fixed_lot_to_decimal(abs(event.data.size)),  # TODO: check this
            liquidation_size=utils.convert_fixed_lot_to_decimal(abs(event.data.size)),
            remaining_liquidatee_balance=utils.convert_fixed_int_to_decimal(event.data.remaining_liquidatee_balance),
            remaining_liquidator_balance=utils.convert_fixed_int_to_decimal(event.data.remaining_liquidator_balance),
            mark_price=utils.convert_fixed_int_to_decimal(event.data.mark_price),
            oracle_price=utils.convert_fixed_int_to_decimal(event.data.underlying_price),
            liquidatee=event.data.liquidatee,
            liquidator=event.data.liquidator,
            asset=Asset.from_index(event.data.asset.index),
            liquidatee_margin_account=event.data.liquidatee_margin_account,
            liquidator_margin_account=event.data.liquidator_margin_account,
        )


@dataclass
class ApplyFundingEvent:
    """Program event for funding being applied to a position."""

    margin_account: Pubkey
    authority: Pubkey
    asset: Asset
    balance_change: float
    remaining_balance: float
    funding_rate: float
    oracle_price: float
    position_size: float

    @classmethod
    def from_event(cls, event: Event):
        assert event.name == cls.__name__
        return cls(
            margin_account=event.data.margin_account,
            authority=event.data.user,
            asset=Asset.from_index(event.data.asset.index),
            balance_change=utils.convert_fixed_int_to_decimal(event.data.balance_change),
            remaining_balance=utils.convert_fixed_int_to_decimal(event.data.remaining_balance),
            funding_rate=utils.convert_fixed_int_to_decimal(event.data.funding_rate),
            oracle_price=utils.convert_fixed_int_to_decimal(event.data.oracle_price),
            position_size=utils.convert_fixed_lot_to_decimal(event.data.position_size),
        )


ZetaEvent = Union[
    PlaceOrderEvent, PlaceMultiOrdersEvent, TradeEvent, CancelOrderEvent, LiquidationEvent, ApplyFundingEvent
]

ZetaEnrichedEvent = Union[
    PlaceOrderEventWithArgs,
    PlaceMultiOrdersEventWithArgs,
    TradeEvent,
    CancelOrderEvent,
    LiquidationEvent,
    ApplyFundingEvent,
]


@dataclass
class EventMeta:
    slot: int
    error: Optional[Any]
    signature: str

    @property
    def is_successful(self) -> bool:
        return self.error is None



from __future__ import annotations

import typing

from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from ..program_id import PROGRAM_ID


class MigrateToCrossMarginAccountAccounts(typing.TypedDict):
    cross_margin_account: Pubkey
    pricing: Pubkey
    authority: Pubkey


def migrate_to_cross_margin_account(
    accounts: MigrateToCrossMarginAccountAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["cross_margin_account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["pricing"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["authority"], is_signer=True, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x9d5kh\xb8\xbdd\xdc"
    encoded_args = b""
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from ..program_id import PROGRAM_ID


class ExpireSeriesArgs(typing.TypedDict):
    settlement_nonce: int


layout = borsh.CStruct("settlement_nonce" / borsh.U8)


def expire_series(
    args: ExpireSeriesArgs,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = []
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"-\xa2ib,\x15\xab\x7f"
    encoded_args = layout.build(
        {
            "settlement_nonce": args["settlement_nonce"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container


class AnchorDecimalJSON(typing.TypedDict):
    flags: int
    hi: int
    lo: int
    mid: int


@dataclass
class AnchorDecimal:
    layout: typing.ClassVar = borsh.CStruct("flags" / borsh.U32, "hi" / borsh.U32, "lo" / borsh.U32, "mid" / borsh.U32)
    flags: int
    hi: int
    lo: int
    mid: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "AnchorDecimal":
        return cls(flags=obj.flags, hi=obj.hi, lo=obj.lo, mid=obj.mid)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"flags": self.flags, "hi": self.hi, "lo": self.lo, "mid": self.mid}

    def to_json(self) -> AnchorDecimalJSON:
        return {"flags": self.flags, "hi": self.hi, "lo": self.lo, "mid": self.mid}

    @classmethod
    def from_json(cls, obj: AnchorDecimalJSON) -> "AnchorDecimal":
        return cls(flags=obj["flags"], hi=obj["hi"], lo=obj["lo"], mid=obj["mid"])



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from solders.sysvar import RENT
from spl.token.constants import TOKEN_PROGRAM_ID

from .. import types
from ..program_id import PROGRAM_ID


class InitializeZetaStateArgs(typing.TypedDict):
    args: types.initialize_state_args.InitializeStateArgs


layout = borsh.CStruct("args" / types.initialize_state_args.InitializeStateArgs.layout)


class InitializeZetaStateAccounts(typing.TypedDict):
    state: Pubkey
    mint_authority: Pubkey
    serum_authority: Pubkey
    treasury_wallet: Pubkey
    referrals_admin: Pubkey
    referrals_rewards_wallet: Pubkey
    usdc_mint: Pubkey
    admin: Pubkey
    secondary_admin: Pubkey


def initialize_zeta_state(
    args: InitializeZetaStateArgs,
    accounts: InitializeZetaStateAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["mint_authority"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["serum_authority"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["treasury_wallet"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["referrals_admin"], is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["referrals_rewards_wallet"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(pubkey=RENT, is_signer=False, is_writable=False),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["usdc_mint"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=accounts["secondary_admin"], is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"D'K\x8e\xbf\x92^\xde"
    encoded_args = layout.build(
        {
            "args": args["args"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container

from . import anchor_decimal


class PricingParametersJSON(typing.TypedDict):
    option_trade_normalizer: anchor_decimal.AnchorDecimalJSON
    future_trade_normalizer: anchor_decimal.AnchorDecimalJSON
    max_volatility_retreat: anchor_decimal.AnchorDecimalJSON
    max_interest_retreat: anchor_decimal.AnchorDecimalJSON
    max_delta: int
    min_delta: int
    min_volatility: int
    max_volatility: int
    min_interest_rate: int
    max_interest_rate: int


@dataclass
class PricingParameters:
    layout: typing.ClassVar = borsh.CStruct(
        "option_trade_normalizer" / anchor_decimal.AnchorDecimal.layout,
        "future_trade_normalizer" / anchor_decimal.AnchorDecimal.layout,
        "max_volatility_retreat" / anchor_decimal.AnchorDecimal.layout,
        "max_interest_retreat" / anchor_decimal.AnchorDecimal.layout,
        "max_delta" / borsh.U64,
        "min_delta" / borsh.U64,
        "min_volatility" / borsh.U64,
        "max_volatility" / borsh.U64,
        "min_interest_rate" / borsh.I64,
        "max_interest_rate" / borsh.I64,
    )
    option_trade_normalizer: anchor_decimal.AnchorDecimal
    future_trade_normalizer: anchor_decimal.AnchorDecimal
    max_volatility_retreat: anchor_decimal.AnchorDecimal
    max_interest_retreat: anchor_decimal.AnchorDecimal
    max_delta: int
    min_delta: int
    min_volatility: int
    max_volatility: int
    min_interest_rate: int
    max_interest_rate: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "PricingParameters":
        return cls(
            option_trade_normalizer=anchor_decimal.AnchorDecimal.from_decoded(obj.option_trade_normalizer),
            future_trade_normalizer=anchor_decimal.AnchorDecimal.from_decoded(obj.future_trade_normalizer),
            max_volatility_retreat=anchor_decimal.AnchorDecimal.from_decoded(obj.max_volatility_retreat),
            max_interest_retreat=anchor_decimal.AnchorDecimal.from_decoded(obj.max_interest_retreat),
            max_delta=obj.max_delta,
            min_delta=obj.min_delta,
            min_volatility=obj.min_volatility,
            max_volatility=obj.max_volatility,
            min_interest_rate=obj.min_interest_rate,
            max_interest_rate=obj.max_interest_rate,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "option_trade_normalizer": self.option_trade_normalizer.to_encodable(),
            "future_trade_normalizer": self.future_trade_normalizer.to_encodable(),
            "max_volatility_retreat": self.max_volatility_retreat.to_encodable(),
            "max_interest_retreat": self.max_interest_retreat.to_encodable(),
            "max_delta": self.max_delta,
            "min_delta": self.min_delta,
            "min_volatility": self.min_volatility,
            "max_volatility": self.max_volatility,
            "min_interest_rate": self.min_interest_rate,
            "max_interest_rate": self.max_interest_rate,
        }

    def to_json(self) -> PricingParametersJSON:
        return {
            "option_trade_normalizer": self.option_trade_normalizer.to_json(),
            "future_trade_normalizer": self.future_trade_normalizer.to_json(),
            "max_volatility_retreat": self.max_volatility_retreat.to_json(),
            "max_interest_retreat": self.max_interest_retreat.to_json(),
            "max_delta": self.max_delta,
            "min_delta": self.min_delta,
            "min_volatility": self.min_volatility,
            "max_volatility": self.max_volatility,
            "min_interest_rate": self.min_interest_rate,
            "max_interest_rate": self.max_interest_rate,
        }

    @classmethod
    def from_json(cls, obj: PricingParametersJSON) -> "PricingParameters":
        return cls(
            option_trade_normalizer=anchor_decimal.AnchorDecimal.from_json(obj["option_trade_normalizer"]),
            future_trade_normalizer=anchor_decimal.AnchorDecimal.from_json(obj["future_trade_normalizer"]),
            max_volatility_retreat=anchor_decimal.AnchorDecimal.from_json(obj["max_volatility_retreat"]),
            max_interest_retreat=anchor_decimal.AnchorDecimal.from_json(obj["max_interest_retreat"]),
            max_delta=obj["max_delta"],
            min_delta=obj["min_delta"],
            min_volatility=obj["min_volatility"],
            max_volatility=obj["max_volatility"],
            min_interest_rate=obj["min_interest_rate"],
            max_interest_rate=obj["max_interest_rate"],
        )



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container

from . import anchor_decimal


class ProductGreeksJSON(typing.TypedDict):
    delta: int
    vega: anchor_decimal.AnchorDecimalJSON
    volatility: anchor_decimal.AnchorDecimalJSON


@dataclass
class ProductGreeks:
    layout: typing.ClassVar = borsh.CStruct(
        "delta" / borsh.U64,
        "vega" / anchor_decimal.AnchorDecimal.layout,
        "volatility" / anchor_decimal.AnchorDecimal.layout,
    )
    delta: int
    vega: anchor_decimal.AnchorDecimal
    volatility: anchor_decimal.AnchorDecimal

    @classmethod
    def from_decoded(cls, obj: Container) -> "ProductGreeks":
        return cls(
            delta=obj.delta,
            vega=anchor_decimal.AnchorDecimal.from_decoded(obj.vega),
            volatility=anchor_decimal.AnchorDecimal.from_decoded(obj.volatility),
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "delta": self.delta,
            "vega": self.vega.to_encodable(),
            "volatility": self.volatility.to_encodable(),
        }

    def to_json(self) -> ProductGreeksJSON:
        return {
            "delta": self.delta,
            "vega": self.vega.to_json(),
            "volatility": self.volatility.to_json(),
        }

    @classmethod
    def from_json(cls, obj: ProductGreeksJSON) -> "ProductGreeks":
        return cls(
            delta=obj["delta"],
            vega=anchor_decimal.AnchorDecimal.from_json(obj["vega"]),
            volatility=anchor_decimal.AnchorDecimal.from_json(obj["volatility"]),
        )



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from spl.token.constants import TOKEN_PROGRAM_ID

from ..program_id import PROGRAM_ID


class InitializeCombinedVaultArgs(typing.TypedDict):
    nonce: int


layout = borsh.CStruct("nonce" / borsh.U8)


class InitializeCombinedVaultAccounts(typing.TypedDict):
    state: Pubkey
    vault: Pubkey
    usdc_mint: Pubkey
    admin: Pubkey


def initialize_combined_vault(
    args: InitializeCombinedVaultArgs,
    accounts: InitializeCombinedVaultAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["vault"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["usdc_mint"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b";ci\x11Iw\xe5\xfc"
    encoded_args = layout.build(
        {
            "nonce": args["nonce"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)
